import argparse
from functools import partial
import glob
import json
import logging
import os
import pathlib
import sys
import time
from typing import Any, List, Optional
from urllib.parse import urlparse

import ansys.pyensight.core

original_stderr = sys.stderr
original_stdout = sys.stdout
sys.stderr = open(os.devnull, "w")
sys.stdout = open(os.devnull, "w")
try:
    import ansys.pyensight.core.utils.dsg_server as dsg_server
    import ansys.pyensight.core.utils.omniverse_dsg_server as ov_dsg_server
    import ansys.pyensight.core.utils.omniverse_glb_server as ov_glb_server
except AttributeError as exc:
    if "_ARRAY_API" not in str(exc):
        raise exc
finally:
    sys.stderr = original_stderr
    sys.stdout = original_stdout


def str2bool_type(v: Any) -> bool:
    """
    This function is designed to be a 'type=' filter for an argparse entry returning a boolean.
    It allows for additional, common alternative strings as booleans.  These include 'yes','no',
    'true','false','t','f','y','n','1' and '0'.  If the value does not meet the requirements,
    the function will raise the argparse.ArgumentTypeError exception.
    :param v: The (potential) boolean argument.
    :return: The actual boolean value.
    :raises: argparse.ArgumentTypeError
    """
    if isinstance(v, bool):
        return v
    if v.lower() in ("yes", "true", "t", "y", "1"):
        return True
    elif v.lower() in ("no", "false", "f", "n", "0"):
        return False
    else:
        raise argparse.ArgumentTypeError("Boolean value expected.")


def int_range_type(
    v: Any, min_value: int = 0, max_value: int = 100, allow: Optional[List[int]] = None
) -> int:
    """
    This function is designed to be a 'type=' filter for an argparse entry returning an integer value within
    a specified range.  If the value does not meet the requirements, the function will raise the
    argparse.ArgumentTypeError exception.  This function is normally used with functools.partial to bind
    the minimum and maximum values.  For example:  type=partial(int_range_type, min_value=0, max_value=65535)
    :param v: The (potential) integer argument.
    :param min_value: The minimum legal integer value.
    :param max_value: The maximum legal integer value.
    :param allow：A list of additional, legal values
    :return: The validated integer value.
    :raises: argparse.ArgumentTypeError
    """
    try:
        value = int(v)
    except ValueError:
        raise argparse.ArgumentTypeError("Integer value expected.")
    if allow is None:
        allow = []
    if (value >= min_value) and (value <= max_value):
        return value
    elif value in allow:
        return value
    else:
        msg = f"Integer value is not in the range [{min_value},{max_value}]"
        if allow:
            msg += f" or in the list {allow}"
        raise argparse.ArgumentTypeError(msg + ".")


class OmniverseGeometryServer(object):
    def __init__(
        self,
        security_token: str = "",
        destination: str = "",
        temporal: bool = False,
        vrmode: bool = False,
        time_scale: float = 1.0,
        normalize_geometry: bool = False,
        dsg_uri: str = "",
        monitor_directory: str = "",
        line_width: float = 0.0,
        grpc_use_tcp_sockets: bool = False,
        grpc_allow_network_connections: bool = False,
        grpc_disable_tls: bool = False,
    ) -> None:
        self._dsg_uri = dsg_uri
        self._destination = destination
        self._security_token = security_token
        if not self._security_token:
            self._security_token = os.environ.get("ENSIGHT_SECURITY_TOKEN", "")
        self._temporal = temporal
        self._vrmode = vrmode
        self._time_scale = time_scale
        self._normalize_geometry = normalize_geometry
        self._version = "unknown"
        self._shutdown = False
        self._server_process = None
        self._status_filename: str = ""
        self._monitor_directory: str = monitor_directory
        self._line_width = line_width
        self._grpc_allow_network_connections = grpc_allow_network_connections
        self._grpc_disable_tls = grpc_disable_tls
        self._grpc_use_tcp_sockets = grpc_use_tcp_sockets

    @property
    def monitor_directory(self) -> Optional[str]:
        if self._monitor_directory:
            return self._monitor_directory
        # converts "" -> None
        return None

    @property
    def pyensight_version(self) -> str:
        """The ansys.pyensight.core version"""
        return ansys.pyensight.core.VERSION

    @property
    def dsg_uri(self) -> str:
        """The endpoint of a Dynamic Scene Graph service:  grpc://{hostname}:{port}"""
        return self._dsg_uri

    @dsg_uri.setter
    def dsg_uri(self, uri: str) -> None:
        self._dsg_uri = uri

    @property
    def destination(self) -> str:
        """The endpoint of an Omniverse Nucleus service:  omniverse://{hostname}/{path}"""
        return self._destination

    @destination.setter
    def destination(self, value: str) -> None:
        self._destination = value

    @property
    def security_token(self) -> str:
        """The security token of the DSG service instance."""
        return self._security_token

    @security_token.setter
    def security_token(self, value: str) -> None:
        self._security_token = value

    @property
    def temporal(self) -> bool:
        """If True, the DSG update should include all timesteps."""
        return self._temporal

    @temporal.setter
    def temporal(self, value: bool) -> None:
        self._temporal = bool(value)

    @property
    def vrmode(self) -> bool:
        """If True, the DSG update should not include camera transforms."""
        return self._vrmode

    @vrmode.setter
    def vrmode(self, value: bool) -> None:
        self._vrmode = bool(value)

    @property
    def normalize_geometry(self) -> bool:
        """If True, the DSG geometry should be remapped into normalized space."""
        return self._normalize_geometry

    @normalize_geometry.setter
    def normalize_geometry(self, val: bool) -> None:
        self._normalize_geometry = val

    @property
    def time_scale(self) -> float:
        """Value to multiply DSG time values by before passing to Omniverse"""
        return self._time_scale

    @time_scale.setter
    def time_scale(self, value: float) -> None:
        self._time_scale = value

    @property
    def line_width(self) -> float:
        return self._line_width

    @line_width.setter
    def line_width(self, line_width: float) -> None:
        self._line_width = line_width

    def run_server(self, one_shot: bool = False) -> None:
        """
        Run a DSG to Omniverse server in process.

        Note: this method does not return until the DSG connection is dropped or
        self.stop_server() has been called.

        Parameters
        ----------
        one_shot : bool
            If True, only run the server to transfer a single scene and
            then return.
        """

        # Build the Omniverse connection
        omni_link = ov_dsg_server.OmniverseWrapper(
            destination=self._destination, line_width=self.line_width
        )
        logging.info("Omniverse connection established.")

        # parse the DSG URI
        uds_path = None
        port = None
        host = None
        if self.dsg_uri.endswith(".sock"):
            # The path is a uds path
            uds_path = self.dsg_uri.replace("unix:", "").replace(".sock", "")
        else:
            parsed = urlparse(self.dsg_uri)
            port = parsed.port
            host = parsed.hostname

        # link it to a DSG session
        update_handler = ov_dsg_server.OmniverseUpdateHandler(omni_link)
        dsg_link = dsg_server.DSGSession(
            port=port,
            host=host,
            vrmode=self.vrmode,
            security_code=self.security_token,
            verbose=1,
            normalize_geometry=self.normalize_geometry,
            time_scale=self.time_scale,
            handler=update_handler,
            uds_path=uds_path,
            grpc_disable_tls=self._grpc_disable_tls,
            grpc_allow_network_connections=self._grpc_allow_network_connections,
            grpc_use_tcp_sockets=self._grpc_use_tcp_sockets,
        )

        # Start the DSG link
        logging.info(f"Making DSG connection to: {self.dsg_uri}")
        err = dsg_link.start()
        if err < 0:
            logging.error("Omniverse connection failed.")
            return

        # Initial pull request
        dsg_link.request_an_update(animation=self.temporal)

        # until the link is dropped, continue
        while not dsg_link.is_shutdown() and not self._shutdown:
            # Reset the line width to the CLI default before each update
            omni_link.line_width = self.line_width
            dsg_link.handle_one_update()

            if one_shot:
                break

        logging.info("Shutting down DSG connection")
        dsg_link.end()
        omni_link.shutdown()

    def run_monitor(self):
        """
        Run monitor and upload GLB files to Omniverse in process.  There are two cases:

        1) the "directory name" is actually a .glb file.  In this case, simply push
        the glb file contents to Omniverse.

        2) If a directory, then we periodically scan the directory for files named "*.upload".
        If this file is found, there are two cases:

            a) The file is empty.  In this case, for a file named ABC.upload, the file
            ABC.glb will be read and uploaded before both files are deleted.

            b) The file contains valid json.  In this case, the json object is parsed with
            the following format (two glb files for the first timestep and one for the second):

                {
                    "version": 1,
                    "destination": "",
                    "files": ["a.glb", "b.glb", "c.glb"],
                    "times": [0.0, 0.0, 1.0]
                }

            "times" is optional and defaults to [0*len("files")].  Once processed,
            all the files referenced in the json and the json file itself are deleted.
            "omniuri" is optional and defaults to the passed Omniverse path.

            Note: In this mode, the method does not return until a "shutdown" file or
            an error is encountered.

            TODO: add "push" mechanism to trigger a DSG push from the connected session.  This
                can be done via the monitor mechanism and used by the Omniverse kit to implement
                a "pull".
        """
        the_dir = self.monitor_directory
        single_file_upload = False
        if os.path.isfile(the_dir) and the_dir.lower().endswith(".glb"):
            single_file_upload = True
        else:
            if not os.path.isdir(the_dir):
                logging.error(f"The monitor directory {the_dir} does not exist.")
                return

        # Build the Omniverse connection
        omni_link = ov_dsg_server.OmniverseWrapper(
            destination=self._destination, line_width=self.line_width
        )
        logging.info("Omniverse connection established.")

        # use an OmniverseUpdateHandler
        update_handler = ov_dsg_server.OmniverseUpdateHandler(omni_link)

        # Link it to the GLB file monitoring service
        glb_link = ov_glb_server.GLBSession(verbose=1, handler=update_handler, vrmode=self.vrmode)
        if single_file_upload:
            start_time = time.time()
            logging.info(f"Uploading file: {the_dir}.")
            try:
                glb_link.start_uploads([0.0, 0.0])
                glb_link.upload_file(the_dir)
                glb_link.end_uploads()
            except Exception as error:
                logging.error(f"Unable to upload file: {the_dir}: {error}")
            logging.info(f"Uploaded in {(time.time() - start_time):.2f}")
        else:
            logging.info(f"Starting file monitoring for {the_dir}.")
            the_dir_path = pathlib.Path(the_dir)
            try:
                stop_file = os.path.join(the_dir, "shutdown")
                orig_destination = omni_link.destination
                while not os.path.exists(stop_file):
                    loop_time = time.time()
                    files_to_remove = []
                    for filename in glob.glob(os.path.join(the_dir, "*.upload")):
                        # reset to the launch URI/directory
                        omni_link.destination = orig_destination
                        # Keep track of the files and time values
                        files_to_remove.append(filename)
                        files_to_process = []
                        file_timestamps = []
                        if os.path.getsize(filename) == 0:
                            # replace the ".upload" extension with ".glb"
                            glb_file = os.path.splitext(filename)[0] + ".glb"
                            if os.path.exists(glb_file):
                                files_to_process.append(glb_file)
                                file_timestamps.append(0.0)
                                files_to_remove.append(glb_file)
                        else:
                            # read the .upload file json content
                            try:
                                with open(filename, "r") as fp:
                                    glb_info = json.load(fp)
                            except Exception:
                                logging.error(f"Unable to read file: {filename}")
                                continue
                            # if specified, set the URI/directory target
                            omni_link.destination = glb_info.get("destination", orig_destination)
                            # Get the GLB files to process
                            the_files = glb_info.get("files", [])
                            files_to_remove.extend(the_files)
                            # Times not used for now, but parse them anyway
                            the_times = glb_info.get("times", [0.0] * len(the_files))
                            file_timestamps.extend(the_times)
                            # Validate a few things
                            if len(the_files) != len(the_times):
                                logging.error(
                                    f"Number of times and files are not the same in: {filename}"
                                )
                                continue
                            files_to_process.extend(the_files)
                        # manage time
                        timeline = sorted(set(file_timestamps))
                        if len(timeline) != 1:
                            logging.warning("Time values not currently supported.")
                        if len(files_to_process) > 1:
                            logging.warning("Multiple glb files not currently fully supported.")
                        # Reset the line width to the CLI default before each update
                        omni_link.line_width = self.line_width
                        # Upload the files
                        glb_link.start_uploads([timeline[0], timeline[-1]])
                        for glb_file, timestamp in zip(files_to_process, file_timestamps):
                            start_time = time.time()
                            logging.info(f"Uploading file: {glb_file} to {omni_link.destination}.")
                            try:
                                time_idx = timeline.index(timestamp) + 1
                                if time_idx == len(timeline):
                                    time_idx -= 1
                                limits = [timestamp, timeline[time_idx]]
                                glb_link.upload_file(glb_file, timeline=limits)
                            except Exception as error:
                                logging.error(f"Unable to upload file: {glb_file}: {error}")
                            logging.info(f"Uploaded in {(time.time() - start_time):.2f}s")
                        glb_link.end_uploads()
                    for filename in files_to_remove:
                        try:
                            # Only delete the file if it is in the_dir_path
                            filename_path = pathlib.Path(filename)
                            if filename_path.is_relative_to(the_dir_path):
                                os.remove(filename)
                        except IOError:
                            pass
                    if time.time() - loop_time < 0.1:
                        time.sleep(0.25)
            except Exception as error:
                logging.error(f"Error encountered while monitoring: {error}")
            logging.info("Stopping file monitoring.")
            try:
                os.remove(stop_file)
            except IOError:
                logging.error("Unable to remove 'shutdown' file.")

        omni_link.shutdown()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="PyEnSight Omniverse Geometry Service")
    parser.add_argument(
        "destination", default="", type=str, help="The directory to save the USD scene graph into."
    )
    parser.add_argument(
        "--verbose",
        metavar="verbose_level",
        default=0,
        type=partial(int_range_type, min_value=0, max_value=3),
        help="Enable logging information (0-3).  Default: 0",
    )
    parser.add_argument(
        "--log_file",
        metavar="log_filename",
        default="",
        type=str,
        help="Save logging output to the named log file instead of stdout.",
    )
    parser.add_argument(
        "--dsg_uri",
        default="grpc://127.0.0.1:5234",
        type=str,
        help="The URI of the EnSight Dynamic Scene Graph server.  Default: grpc://127.0.0.1:5234",
    )
    parser.add_argument(
        "--security_token",
        metavar="token",
        default="",
        type=str,
        help="Dynamic scene graph API security token.  Default: none",
    )
    parser.add_argument(
        "--monitor_directory",
        metavar="glb_directory",
        default="",
        type=str,
        help="Monitor specified directory for GLB files to be exported.  Default: none",
    )
    parser.add_argument(
        "--time_scale",
        metavar="time_scale",
        default=1.0,
        type=float,
        help="Scaling factor to be applied to input time values.  Default: 1.0",
    )
    parser.add_argument(
        "--normalize_geometry",
        metavar="yes|no|true|false|1|0",
        default=False,
        type=str2bool_type,
        help="Enable mapping of geometry to a normalized Cartesian space. Default: false",
    )
    parser.add_argument(
        "--include_camera",
        metavar="yes|no|true|false|1|0",
        default=True,
        type=str2bool_type,
        help="Include the camera in the output USD scene graph. Default: true",
    )
    parser.add_argument(
        "--temporal",
        metavar="yes|no|true|false|1|0",
        default=False,
        type=str2bool_type,
        help="Export a temporal scene graph. Default: false",
    )
    parser.add_argument(
        "--oneshot",
        metavar="yes|no|true|false|1|0",
        default=False,
        type=str2bool_type,
        help="Convert a single geometry into USD and exit.  Default: false",
    )
    line_default: Any = os.environ.get("ANSYS_OV_LINE_WIDTH", None)
    if line_default is not None:
        try:
            line_default = float(line_default)
        except ValueError:
            line_default = None
    parser.add_argument(
        "--line_width",
        metavar="line_width",
        default=line_default,
        type=float,
        help=f"Width of lines: >0=absolute size. <0=fraction of diagonal. 0=none. Default: {line_default}",
    )
    parser.add_argument(
        "--grpc_use_tcp_sockets",
        metavar="yes|no|true|false|1|0",
        default=False,
        type=str2bool_type,
        help="If using gRPC, and if True, then allow TCP Socket based connections instead of only local connections.",
    )
    parser.add_argument(
        "--grpc_allow_network_connections",
        metavar="yes|no|true|false|1|0",
        default=False,
        type=str2bool_type,
        help="If using gRPC and using TCP Socket based connections, listen on all networks.",
    )
    parser.add_argument(
        "--grpc_disable_tls",
        metavar="yes|no|true|false|1|0",
        default=False,
        type=str2bool_type,
        help="If using gRPC and using TCP Socket based connections, disable TLS.",
    )
    # parse the command line
    args = parser.parse_args()

    # set up logging
    level = logging.ERROR
    if args.verbose == 1:
        level = logging.WARN
    elif args.verbose == 2:
        level = logging.INFO
    elif args.verbose == 3:
        level = logging.DEBUG
    log_args = dict(format="GeometryService:%(levelname)s:%(message)s", level=level)
    if args.log_file:
        log_args["filename"] = args.log_file
    # start with a clean logging instance
    while logging.root.hasHandlers():
        logging.root.removeHandler(logging.root.handlers[0])
    logging.basicConfig(**log_args)  # type: ignore

    # size of lines in data units or fraction of bounding box diagonal
    line_width = 0.0
    if args.line_width is not None:
        line_width = args.line_width

    # Build the server object
    server = OmniverseGeometryServer(
        destination=args.destination,
        dsg_uri=args.dsg_uri,
        security_token=args.security_token,
        monitor_directory=args.monitor_directory,
        time_scale=args.time_scale,
        normalize_geometry=args.normalize_geometry,
        vrmode=not args.include_camera,
        temporal=args.temporal,
        line_width=line_width,
        grpc_disable_tls=args.grpc_disable_tls,
        grpc_allow_network_connections=args.grpc_allow_network_connections,
        grpc_use_tcp_sockets=args.grpc_use_tcp_sockets,
    )

    # run the server
    logging.info("Server startup.")
    if server.monitor_directory:
        server.run_monitor()
    else:
        server.run_server(one_shot=args.oneshot)
    logging.info("Server shutdown.")



"""Views module.

The Views module allows PyEnSight to control the view in the EnSight session.

Example to set an isometric view:

::
    (PyEnSight)
    from ansys.pyensight.core import LocalLauncher
    session = LocalLauncher().start()
    views = session.ensight.utils.views
    views.set_view_direction(1,1,1)

    (EnSight)
    from ensight.utils import views
    views.set_view_direction(1,1,1)


"""


import math
from types import ModuleType
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Union

import numpy as np

if TYPE_CHECKING:
    try:
        import ensight
    except ImportError:
        from ansys.api.pyensight import ensight_api


VIEW_DICT = {
    "x+": (1, 0, 0),
    "x-": (-1, 0, 0),
    "y+": (0, 1, 0),
    "y-": (0, -1, 0),
    "z+": (0, 0, 1),
    "z-": (0, 0, -1),
    "isometric": (1, 1, 1),
}


class _Simba:
    """Hidden class to manage the interactor layer in simba"""

    def __init__(self, ensight: Union["ensight_api.ensight", "ensight"], views: "Views"):
        self.ensight = ensight
        self.views = views
        self._original_look_at = None
        self._original_look_from = None
        self._original_parallel_scale = None
        self._original_view_angle = None
        self._original_view_up = None

    def _initialize_simba_view(self):
        """Initialize the data for resetting the camera."""
        vport = self.ensight.objs.core.VPORTS[0]
        near_clip = vport.ZCLIPLIMITS[0]
        view_angle = 2 * vport.PERSPECTIVEANGLE
        self._original_parallel_scale = near_clip * math.tan(math.radians(view_angle) / 2)
        self._original_view_angle = view_angle
        (
            self._original_look_from,
            self._original_look_at,
            self._original_view_up,
            self._original_parallel_scale,
        ) = self.compute_camera_from_ensight_opengl()
        self.ensight.annotation.axis_global("off")
        self.ensight.annotation.axis_local("off")
        self.ensight.annotation.axis_model("off")
        self.ensight.view_transf.zclip_float("OFF")

    def get_center_of_rotation(self):
        """Get EnSight center of rotation."""
        return self.ensight.objs.core.VPORTS[0].TRANSFORMCENTER.copy()

    def auto_scale(self):
        """Auto scale view."""
        vport = self.ensight.objs.core.VPORTS[0]
        vport.PERSPECTIVEANGLE = 14  # on fit we need to set a default angle.
        self.ensight.view_transf.function("global")
        self.ensight.view_transf.fit()
        self._initialize_simba_view()
        self.render()
        return self.get_camera()

    def set_view(self, value: str):
        """Set the view."""
        self.ensight.view_transf.function("global")
        if value != "isometric":
            new_value = value[1].upper() + value[0]
            self.ensight.view_transf.view_recall(new_value)
        else:
            self.views.set_view_direction(
                1, 1, 1, perspective=self.ensight.objs.core.vports[0].PERSPECTIVE
            )
        return self.auto_scale()

    def get_camera(self):
        """Get EnSight camera settings in VTK format."""
        vport = self.ensight.objs.core.VPORTS[0]
        position, focal_point, view_up, parallel_scale = self.compute_camera_from_ensight_opengl()
        vport = self.ensight.objs.core.VPORTS[0]
        view_angle = 2 * vport.PERSPECTIVEANGLE
        # The parameter parallel scale is the actual parallel scale only
        # if the vport is in orthographic mode. If not, it is defined as the
        # inverge of the tangent of half of the field of view
        parallel_scale = parallel_scale
        return {
            "orthographic": not vport.PERSPECTIVE,
            "view_up": view_up,
            "position": position,
            "focal_point": focal_point,
            "view_angle": view_angle,
            "parallel_scale": parallel_scale,
            "reset_focal_point": self._original_look_at,
            "reset_position": self._original_look_from,
            "reset_parallel_scale": self._original_parallel_scale,
            "reset_view_up": self._original_view_up,
            "reset_view_angle": self._original_view_angle,
        }

    @staticmethod
    def normalize(v):
        """Normalize a numpy vector."""
        norm = np.linalg.norm(v)
        return v / norm if norm > 0 else v

    @staticmethod
    def rotation_matrix_to_quaternion(m):
        """Convert a numpy rotation matrix to a quaternion."""
        trace = np.trace(m)
        if trace > 0:
            s = 0.5 / np.sqrt(trace + 1.0)
            w = 0.25 / s
            x = (m[2, 1] - m[1, 2]) * s
            y = (m[0, 2] - m[2, 0]) * s
            z = (m[1, 0] - m[0, 1]) * s
        else:
            if m[0, 0] > m[1, 1] and m[0, 0] > m[2, 2]:
                s = 2.0 * np.sqrt(1.0 + m[0, 0] - m[1, 1] - m[2, 2])
                w = (m[2, 1] - m[1, 2]) / s
                x = 0.25 * s
                y = (m[0, 1] + m[1, 0]) / s
                z = (m[0, 2] + m[2, 0]) / s
            elif m[1, 1] > m[2, 2]:
                s = 2.0 * np.sqrt(1.0 + m[1, 1] - m[0, 0] - m[2, 2])
                w = (m[0, 2] - m[2, 0]) / s
                x = (m[0, 1] + m[1, 0]) / s
                y = 0.25 * s
                z = (m[1, 2] + m[2, 1]) / s
            else:
                s = 2.0 * np.sqrt(1.0 + m[2, 2] - m[0, 0] - m[1, 1])
                w = (m[1, 0] - m[0, 1]) / s
                x = (m[0, 2] + m[2, 0]) / s
                y = (m[1, 2] + m[2, 1]) / s
                z = 0.25 * s
        return np.array([x, y, z, w])

    def compute_camera_from_ensight_opengl(self):
        """Simulate a rotating camera using the current quaternion."""
        if isinstance(self.ensight, ModuleType):
            data = self.ensight.objs.core.VPORTS[0].simba_camera()
        else:
            data = self.ensight._session.cmd("ensight.objs.core.VPORTS[0].simba_camera())")
        camera_position = [data[0], data[1], data[2]]
        focal_point = [data[3], data[4], data[5]]
        view_up = [data[6], data[7], data[8]]
        parallel_scale = 1 / data[9]
        return camera_position, focal_point, self.views._normalize_vector(view_up), parallel_scale

    def set_camera(
        self,
        orthographic,
        view_up=None,
        position=None,
        focal_point=None,
        view_angle=None,
    ):
        """Set the EnSight camera settings from the VTK input."""
        self.ensight.view_transf.function("global")
        perspective = "OFF" if orthographic else "ON"
        self.ensight.view.perspective(perspective)
        vport = self.ensight.objs.core.VPORTS[0]
        vport.PERSPECTIVEANGLE = view_angle / 2
        if view_up and position and focal_point:
            current_camera = self.get_camera()
            center = vport.TRANSFORMCENTER.copy()
            if isinstance(self.ensight, ModuleType):
                data = self.ensight.objs.core.VPORTS[0].simba_set_camera_helper(
                    position,
                    focal_point,
                    view_up,
                    current_camera["position"],
                    current_camera["focal_point"],
                    current_camera["view_up"],
                    center,
                    vport.ROTATION.copy(),
                )
            else:
                current_position = current_camera["position"]
                current_fp = current_camera["focal_point"]
                current_up = current_camera["view_up"]
                cmd = "ensight.objs.core.VPORTS[0].simba_set_camera_helper("
                cmd += f"{position}, {focal_point}, {view_up}, {current_position}, "
                cmd += f"{current_fp}, {current_up}, {center}, "
                cmd += f"{vport.ROTATION.copy()})"
                data = self.ensight._session.cmd(cmd)
            self.ensight.view_transf.rotate(data[3], data[4], data[5])
            self.ensight.view_transf.translate(data[0], data[1], 0)

        self.render()
        return self.get_camera()

    def set_perspective(self, value):
        self.ensight.view_transf.function("global")
        vport = self.ensight.objs.core.VPORTS[0]
        self.ensight.view.perspective(value)
        vport.PERSPECTIVE = value == "ON"
        self.ensight.view_transf.zoom(1)
        self.ensight.view_transf.rotate(0, 0, 0)
        self.render()
        return self.get_camera()

    def screen_to_world(self, mousex, mousey, invert_y=False, set_center=False):
        mousex = int(mousex)
        mousey = int(mousey)
        if isinstance(self.ensight, ModuleType):
            model_point = self.ensight.objs.core.VPORTS[0].screen_to_coords(
                mousex, mousey, invert_y, set_center
            )
        else:
            model_point = self.ensight._session.cmd(
                f"ensight.objs.core.VPORTS[0].screen_to_coords({mousex}, {mousey}, {invert_y}, {set_center})"
            )
        return {"model_point": model_point, "camera": self.get_camera()}

    def render(self):
        """Force render update in EnSight."""
        self.ensight.refresh()

    def _probe_setup(self, part_obj, get_probe_data=False):
        self.ensight.query_interact.number_displayed(100)
        self.ensight.query_interact.query("surface")
        self.ensight.query_interact.display_id("OFF")
        self.ensight.query_interact.label_always_on_top("ON")
        self.ensight.query_interact.marker_size_normalized(2)
        if get_probe_data:
            variable_string = """Coordinates 'X' 'Y' 'Z'"""
            variable_list = [variable_string]
            variable_name = part_obj.COLORBYPALETTE
            if variable_name:
                if isinstance(variable_name, str):
                    variable_list.append(variable_name)
                else:
                    if isinstance(variable_name, list):
                        if variable_name[0]:
                            variable_name = variable_name[0].DESCRIPTION
                            variable_list.append(variable_name)
                        else:
                            variable_name = None
            if isinstance(self.ensight, ModuleType):
                self.ensight.query_interact.select_varname_begin(*variable_list)
            else:
                command = "ensight.query_interact.select_varname_begin("
                for var in variable_list:
                    command += var + ","
                command = command[:-1] + ")"
                self.ensight._session.cmd(command)
        self.render()

    def drag_allowed(self, mousex, mousey, invert_y=False, probe=False, get_probe_data=False):
        """Return True if the picked object is allowed dragging in the interactor."""
        mousex = int(mousex)
        mousey = int(mousey)
        if isinstance(self.ensight, ModuleType):
            part_id, tool_id = self.ensight.objs.core.VPORTS[0].simba_what_is_picked(
                mousex, mousey, invert_y
            )
        else:
            part_id, tool_id = self.ensight._session.cmd(
                f"ensight.objs.core.VPORTS[0].simba_what_is_picked({mousex}, {mousey}, {invert_y})"
            )
        coords = [None, None, None]
        if probe:
            screen_to_world = self.screen_to_world(
                mousex=mousex, mousey=mousey, invert_y=invert_y, set_center=False
            )
            coords = screen_to_world["model_point"]
        if tool_id > -1:
            return True, coords[0], coords[1], coords[2], False
        part_types_allowed = [
            self.ensight.objs.enums.PART_CLIP_PLANE,
            self.ensight.objs.enums.PART_ISO_SURFACE,
            self.ensight.objs.enums.PART_CONTOUR,
        ]
        if part_id > -1:
            part_obj = self.ensight.objs.core.PARTS.find(part_id, "PARTNUMBER")[0]
            if probe:
                width, height = tuple(self.ensight.objs.core.WINDOWSIZE)
                if invert_y:
                    mousey = height - mousey
                self.ensight.query_interact.number_displayed(100)
                self.ensight.query_interact.query("surface")
                self.ensight.query_interact.display_id("OFF")
                self.ensight.query_interact.create(mousex / width, mousey / height)
                self._probe_setup(part_obj, get_probe_data=get_probe_data)
            return part_obj.PARTTYPE in part_types_allowed, coords[0], coords[1], coords[2], True
        if (
            get_probe_data and self.ensight.objs.core.PROBES[0].PROBE_DATA
        ):  # In case we have picked a probe point
            for part in self.ensight.objs.core.PARTS:
                self._probe_setup(part, get_probe_data=get_probe_data)
        return False, coords[0], coords[1], coords[2], False


class Views:
    """Controls the view in the current EnSight ``Session`` instance."""

    def __init__(self, ensight: Union["ensight_api.ensight", "ensight"]):
        self.ensight = ensight
        self._views_dict: Dict[str, Tuple[int, List[float]]] = {}
        self._simba = _Simba(ensight, self)

    @staticmethod
    def _normalize_vector(direction: List[float]) -> List[float]:
        """Return the normalized input (3D) vector.

        Parameters
        ----------
        direction : list[float]
            List representing the vector to normalize.

        Returns
        -------
        list[float]
            List representing the normalized vector.

        """
        magnitude = math.sqrt(sum(v**2 for v in direction))
        if magnitude == 0.0:
            return [0] * len(direction)
        return [x / magnitude for x in direction]

    @staticmethod
    def _cross_product(vec1: List[float], vec2: List[float]) -> List[float]:
        """Get the cross product of two input vectors.

        Parameters
        ----------
        vec1 : list[float]
            List representing the first vector.
        vec2 : list[float]
            List representing the second vector.

        Returns
        -------
        list[float]
            List representing the cross product of the two input vectors.

        """
        return [
            vec1[1] * vec2[2] - vec1[2] * vec2[1],
            vec1[2] * vec2[0] - vec1[0] * vec2[2],
            vec1[0] * vec2[1] - vec1[1] * vec2[0],
        ]

    def _convert_view_direction_to_rotation_matrix(
        self, direction: List[float], up_axis: Tuple[float, float, float] = (0.0, 1.0, 0.0)
    ) -> Tuple[List[float], List[float], List[float]]:
        """Convert the input direction vector in a rotation matrix.

        The third row of the rotation matrix is the view direction.

        The first and second rows are computed to be orthogonal to the view direction,
        forming an orthogonal matrix. To get a specific view direction, a rotation matrix has
        the third column, which is the view direction itself (that is the view direction becomes the z
        axis after the rotation, while the transformed x and y axis are computed to be orthogonal to
        the z transformed axis). The rotation is defined as the matrix transpose of the
        defined rotation matrix because the aim is to have the view direction pointing towards the camera
        and not the contrary.

        Parameters
        ----------
        direction : list[float]
            List describing the desired direction view
        up_axis : tuple[float, float, float], optional
            Tuple describing the up direction. The default is ``(0.0, 1.0, 0.0)``,
            which assumes the Y axis.

        Returns
        -------
        tuple
            Tuple containing the three rows of the rotation matrix.

        """
        direction = self._normalize_vector(direction)
        xaxis = self._normalize_vector(self._cross_product(list(up_axis), direction))
        yaxis = self._normalize_vector(self._cross_product(direction, xaxis))
        return (xaxis, yaxis, direction)

    def _convert_view_direction_to_quaternion(
        self, direction: List[float], up_axis: Tuple[float, float, float] = (0.0, 1.0, 0.0)
    ) -> Tuple[float, float, float, float]:
        """Convert the input direction vector into a list of quaternions.

        Parameters
        ----------
        direction : list
            List describing the desired direction view.
        up_axis : tuple
            Tuple describing the up direction. The default is ``(0.0, 1.0, 0.0)``,
            which assumes the Y axis.

        Returns
        -------
        tuple
            Tuple containing the four quaternions describing the required rotation.

        """
        row0, row1, row2 = self._convert_view_direction_to_rotation_matrix(
            direction=direction,
            up_axis=up_axis,
        )
        return self._convert_rotation_matrix_to_quaternion(row0, row1, row2)

    def _convert_rotation_matrix_to_quaternion(
        self, row0: List[float], row1: List[float], row2: List[float]
    ) -> Tuple[float, float, float, float]:
        """Convert a rotation matrix to quaternions.

        Parameters
        ----------
        row0 : list[float]
            First row of the matrix.
        row1 : list[float]
            Second row of the matrix.
        row2 : list[float]
            Third row of the matrix.

        Returns
        -------
        tuple
            Four quaternions describing the rotation.

        """
        trace = row0[0] + row1[1] + row2[2]
        if trace > 0:
            s = math.sqrt(trace + 1)
            qw = s / 2
            s = 1 / (2 * s)
            qx = (row2[1] - row1[2]) * s
            qy = (row0[2] - row2[0]) * s
            qz = (row1[0] - row0[1]) * s
        elif row0[0] > row1[1] and row0[0] > row2[2]:
            s = math.sqrt(1 + row0[0] - row1[1] - row2[2])
            qx = s / 2
            s = 1 / (2 * s)
            qw = (row2[1] - row1[2]) * s
            qy = (row0[1] + row1[0]) * s
            qz = (row0[2] + row2[0]) * s
        elif row1[1] > row2[2]:
            s = math.sqrt(1 + row1[1] - row0[0] - row2[2])
            qy = s / 2
            s = 1 / (2 * s)
            qw = (row0[2] - row2[0]) * s
            qx = (row0[1] + row1[0]) * s
            qz = (row1[2] + row2[1]) * s
        else:
            s = math.sqrt(1 + row2[2] - row0[0] - row1[1])
            qz = s / 2
            if s != 0.0:
                s = 1 / (2 * s)
            qw = (row1[0] - row0[1]) * s
            qx = (row0[2] + row2[0]) * s
            qy = (row1[2] + row2[1]) * s
        list_of_quats = self._normalize_vector([qx, qy, qz, qw])
        return list_of_quats[0], list_of_quats[1], list_of_quats[2], list_of_quats[3]

    @property
    def views_dict(self) -> Dict[str, Tuple[int, List[float]]]:
        """Dictionary holding the stored views.

        Returns
        -------
        dict
            Dictionary containing the stored views.

        """
        return self._views_dict

    # Methods
    def set_center_of_transform(self, xc: float, yc: float, zc: float) -> None:
        """Change the center of transform of the current session.

        Parameters
        ----------
        xc : float
            x coordinate of the new center of the transform.
        yc : float
            y coordinate of the new center of the transform.
        zc : float
            z coordinate of the new center of the transform.

        """
        self.ensight.view_transf.center_of_transform(xc, yc, zc)

    def compute_model_centroid(self, vportindex: int = 0) -> List[float]:
        """Compute the model centroid using the model bounds.

        Parameters
        ----------
        vportindex : int, optional
            Viewport to compute the centroid for. The default is ``0``.

        Returns
        -------
        list
            Coordinates of the model centroid.

        """
        vport = self.ensight.objs.core.VPORTS[vportindex]
        try:
            # Available from release 24.1. The order is:
            # xmin,ymin,zmin,xmax,ymax,zmax
            bounds = vport.BOUNDINGBOX
            xmax = bounds[3]
            xmin = bounds[0]
            ymax = bounds[4]
            ymin = bounds[1]
            zmax = bounds[5]
            zmin = bounds[2]
        except AttributeError:  # pragma: no cover
            # Old method. It assumes autosize is set to True and
            # that the bounds have not been modified
            enabled = False  # pragma: no cover
            if self.ensight.objs.core.BOUNDS is False:  # pragma: no cover
                enabled = True  # pragma: no cover
                self.ensight.view.bounds("ON")  # pragma: no cover
            xmax = vport.AXISXMAX  # pragma: no cover
            xmin = vport.AXISXMIN  # pragma: no cover
            ymax = vport.AXISYMAX  # pragma: no cover
            ymin = vport.AXISYMIN  # pragma: no cover
            zmax = vport.AXISZMAX  # pragma: no cover
            zmin = vport.AXISZMIN  # pragma: no cover
            if enabled:  # pragma: no cover
                self.ensight.view.bounds("OFF")  # pragma: no cover
        xavg = (xmax + xmin) / 2
        yavg = (ymax + ymin) / 2
        zavg = (zmax + zmin) / 2
        values = [xavg, yavg, zavg]
        return values

    def set_view_direction(
        self,
        xdir: float,
        ydir: float,
        zdir: float,
        name: Optional[str] = None,
        perspective: Optional[bool] = False,
        up_axis: Tuple[float, float, float] = (0.0, 1.0, 0.0),
        vportindex: int = 0,
    ) -> None:
        """Set the view direction of the session.

        Parameters
        ----------
        xdir : float
            x component of the view direction.
        ydir : float
            y component of the view direction.
        zdir : float
            z component of the view direction.
        name : str, optional
            Name for the new view settings. The default is ``None``,
            in which case an incremental name is automatically assigned.
        perspective : bool, optional
            Whether to enable the perspective view. The default is ``False``.
        up_axis : tuple[float, float, float], optional
            Up direction for the view direction. The default is ``(0.0, 1.0, 0.0)``.
        vportindex : int, optional
            Viewport to set the view direction for. The default is ``0``.
        """
        vport = self.ensight.objs.core.VPORTS[vportindex]
        if not perspective:
            self.ensight.view.perspective("OFF")
            vport.PERSPECTIVE = False
        direction = [xdir, ydir, zdir]
        rots = vport.ROTATION.copy()
        rots[0:4] = self._convert_view_direction_to_quaternion(direction, up_axis=up_axis)
        vport.ROTATION = rots
        if perspective:
            self.ensight.view.perspective("ON")
            vport.PERSPECTIVE = True
        self.save_current_view(name=name, vportindex=vportindex)

    def save_current_view(
        self,
        name: Optional[str] = None,
        vportindex: int = 0,
    ) -> None:
        """Save the current view.

        Parameters
        ----------
        name : str, optional
            Name to give to the new view. The default is ``None``,
            in which case an incremental name is automatically assigned.
        vportindex : int, optional
            Viewport to set the view direction for. The default is ``0``.
        """
        vport = self.ensight.objs.core.VPORTS[vportindex]
        coretransform = vport.CORETRANSFORM
        if not name:
            count = 0
            while True and count < 100:
                if self.views_dict.get("view_{}".format(count)):
                    count += 1
                else:
                    self.views_dict["view_{}".format(count)] = (vportindex, coretransform)
                    break
        else:
            self.views_dict[name] = (vportindex, coretransform)

    def restore_view(self, name: str) -> None:
        """Restore a stored view.

        Parameters
        ----------
        name : str
            Name of the view to restore.

        """
        if not self.views_dict.get(name):
            raise KeyError("ERROR: view set not available")
        found = self.views_dict.get(name)
        if found:  # pragma: no cover
            viewport, coretransform = found
            vport = self.ensight.objs.core.VPORTS[viewport]
            vport.CORETRANSFORM = coretransform

    def restore_center_of_transform(self) -> None:
        """Restore the center of the transform to the model centroid."""
        original_model_centroid = self.compute_model_centroid()
        self.set_center_of_transform(*original_model_centroid)

    def reinitialize_view(self) -> None:
        """Reset the view."""
        self.ensight.view_transf.initialize_viewports()



#
# This file borrows heavily from the Omniverse Example Connector which
# contains the following notice:
#
###############################################################################
# Copyright 2020 NVIDIA Corporation
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#
###############################################################################
import logging
import math
import os
import platform
import shutil
import sys
import tempfile
from typing import Any, Dict, List, Optional
import warnings

from ansys.pyensight.core.utils.dsg_server import Part, UpdateHandler
import numpy
import png

try:
    from pxr import Gf, Kind, Sdf, Usd, UsdGeom, UsdLux, UsdShade
except ModuleNotFoundError:
    if sys.version_info.minor >= 14:
        warnings.warn("USD Export not supported for Python >= 3.14")
        sys.exit(1)
    is_linux_arm64 = platform.system() == "Linux" and platform.machine() == "aarch64"
    if is_linux_arm64:
        warnings.warn("USD Export not supported on Linux ARM platforms")
        sys.exit(1)


class OmniverseWrapper(object):
    def __init__(
        self,
        live_edit: bool = False,
        destination: str = "",
        line_width: float = 0.0,
    ) -> None:
        # File extension.  For debugging, .usda is sometimes helpful.
        self._ext = ".usd"
        self._cleaned_index = 0
        self._cleaned_names: dict = {}
        self._connectionStatusSubscription = None
        self._stage = None
        self._destinationPath: str = ""
        self._old_stages: list = []
        self._stagename: str = "dsg_scene" + self._ext
        self._live_edit: bool = live_edit
        if self._live_edit:
            self._stagename = "dsg_scene.live"
        # USD time slider will have 120 tick marks per second of animation time
        self._time_codes_per_second: float = 120.0
        # Omniverse content currently only scales correctly for scenes in cm.  DJB, Feb 2025
        self._units_per_meter: float = 100.0
        self._up_axis: str = UsdGeom.Tokens.y
        if destination:
            self.destination = destination

        self._line_width = line_width
        self._centroid: Optional[list] = None
        # Record the files per timestep, per mesh type.  {part_name: {"surfaces": [], "lines": [], "points": []} }
        self._time_files: dict = {}

    @property
    def destination(self) -> str:
        """The current output directory."""
        return self._destinationPath

    @destination.setter
    def destination(self, directory: str) -> None:
        self._destinationPath = directory
        if not self.is_valid_destination(directory):
            logging.warning(f"Invalid destination path: {directory}")

    @property
    def line_width(self) -> float:
        return self._line_width

    @line_width.setter
    def line_width(self, line_width: float) -> None:
        self._line_width = line_width

    def shutdown(self) -> None:
        """
        Shutdown the connection to Omniverse cleanly.
        """
        self._connectionStatusSubscription = None

    @staticmethod
    def is_valid_destination(path: str) -> bool:
        """
        Verify that the target path is a writeable directory.

        Parameters
        ----------
        path
            The path to check

        Returns
        -------
            True if the path is a writeable directory, False otherwise.
        """
        return os.access(path, os.W_OK)

    def stage_url(self, name: Optional[str] = None) -> str:
        """
        For a given object name, create the URL for the item.
        Parameters
        ----------
        name: the name of the object to generate the URL for. If None, it will be the URL for the
              stage name.

        Returns
        -------
        The URL for the object.
        """
        if name is None:
            name = self._stagename
        return os.path.join(self._destinationPath, name)

    def delete_old_stages(self) -> None:
        """
        Remove all the stages included in the "_old_stages" list.
        If a stage is in use and cannot be removed, keep its name in _old_stages
        to retry later.
        """
        stages_unremoved = list()
        while self._old_stages:
            stage = self._old_stages.pop()
            try:
                if os.path.isfile(stage):
                    os.remove(stage)
                else:
                    shutil.rmtree(stage, ignore_errors=True, onerror=None)
            except OSError:
                if not stage.endswith("_manifest" + self._ext):
                    stages_unremoved.append(stage)
        self._old_stages = stages_unremoved

    def create_new_stage(self) -> None:
        """
        Create a new stage. using the current stage name.
        """
        logging.info(f"Creating Omniverse stage: {self.stage_url()}")
        if self._stage:
            self._stage.Unload()
            self._stage = None
        self.delete_old_stages()
        self._stage = Usd.Stage.CreateNew(self.stage_url())
        # record the stage in the "_old_stages" list.
        self._old_stages.append(self.stage_url())
        UsdGeom.SetStageUpAxis(self._stage, self._up_axis)
        UsdGeom.SetStageMetersPerUnit(self._stage, 1.0 / self._units_per_meter)
        logging.info(f"Created stage: {self.stage_url()}")

    def save_stage(self, comment: str = "") -> None:
        """
        For live connections, save the current edit and allow live processing.

        Presently, live connections are disabled.
        """
        self._stage.GetRootLayer().Save()  # type:ignore

    def clear_cleaned_names(self) -> None:
        """
        Clear the list of cleaned names
        """
        self._cleaned_names = {}
        self._cleaned_index = 0

    def clean_name(self, name: str, id_name: Any = None) -> str:
        """Generate a valid USD name

        From a base (EnSight) varname, partname, etc. and the DSG id, generate
        a unique, valid USD name.  Save the names so that if the same name
        comes in again, the previously computed name is returned and if the
        manipulation results in a conflict, the name can be made unique.

        Parameters
        ----------
        name:
            The name to generate a USD name for.

        id_name:
            The DSG id associated with the DSG name, if any.

        Returns
        -------
            A unique USD name.
        """
        orig_name = name
        # return any previously generated name
        if (name, id_name) in self._cleaned_names:
            return self._cleaned_names[(name, id_name)]
        # replace invalid characters.  EnSight uses a number of characters that are illegal in USD names.
        replacements = {
            ord("+"): "_",
            ord("-"): "_",
            ord("."): "_",
            ord(":"): "_",
            ord("["): "_",
            ord("]"): "_",
            ord("("): "_",
            ord(")"): "_",
            ord("<"): "_",
            ord(">"): "_",
            ord("/"): "_",
            ord("="): "_",
            ord(","): "_",
            ord(" "): "_",
            ord("\\"): "_",
            ord("^"): "_",
            ord("!"): "_",
            ord("#"): "_",
            ord("%"): "_",
            ord("&"): "_",
        }
        name = name.translate(replacements)
        if name[0].isdigit():
            name = f"_{name}"
        if id_name is not None:
            name = name + "_" + str(id_name)
        if name in self._cleaned_names.values():
            # Make the name unique
            while f"{name}_{self._cleaned_index}" in self._cleaned_names.values():
                self._cleaned_index += 1
            name = f"{name}_{self._cleaned_index}"
        # store off the cleaned name
        self._cleaned_names[(orig_name, id_name)] = name
        return name

    @staticmethod
    def decompose_matrix(values: Any) -> Any:
        """
        Decompose an array of floats (representing a 4x4 matrix) into scale, rotation and translation.
        Parameters
        ----------
        values:
            16 values (input to Gf.Matrix4f CTOR)

        Returns
        -------
        (scale, rotation, translation)
        """
        # ang_convert = 180.0/math.pi
        ang_convert = 1.0
        trans_convert = 1.0
        m = Gf.Matrix4f(*values)
        m = m.GetTranspose()

        s = math.sqrt(m[0][0] * m[0][0] + m[0][1] * m[0][1] + m[0][2] * m[0][2])
        # cleanup scale
        m = m.RemoveScaleShear()
        # r = m.ExtractRotation()
        R = m.ExtractRotationMatrix()
        r = [
            math.atan2(R[2][1], R[2][2]) * ang_convert,
            math.atan2(-R[2][0], 1.0) * ang_convert,
            math.atan2(R[1][0], R[0][0]) * ang_convert,
        ]
        t = m.ExtractTranslation()
        t = [t[0] * trans_convert, t[1] * trans_convert, t[2] * trans_convert]
        return s, r, t

    # Common code to create the part manifest file and the file per timestep
    def create_dsg_surfaces_file(
        self,
        file_url,  # SdfPath, location on disk
        part_path: str,  # base path name, such as "/Root/Case_1/Isosurface_part"
        verts,
        normals,
        conn,
        tcoords,
        diffuse,
        variable,
        mat_info,
        is_manifest: bool,
    ):
        if is_manifest and file_url in self._old_stages:
            return False
        if not is_manifest and os.path.exists(file_url):
            return False

        stage = Usd.Stage.CreateNew(file_url)
        UsdGeom.SetStageUpAxis(stage, self._up_axis)
        UsdGeom.SetStageMetersPerUnit(stage, 1.0 / self._units_per_meter)
        self._old_stages.append(file_url)

        part_prim = stage.OverridePrim(part_path)

        surfaces_prim = self.create_xform_node(stage, part_path + "/surfaces")
        mesh = UsdGeom.Mesh.Define(stage, str(surfaces_prim.GetPath()) + "/Mesh")
        mesh.CreateDoubleSidedAttr().Set(True)
        pt_attr = mesh.CreatePointsAttr()
        if verts is not None:
            pt_attr.Set(verts, 0)
        norm_attr = mesh.CreateNormalsAttr()
        if normals is not None:
            norm_attr.Set(normals, 0)
        fvc_attr = mesh.CreateFaceVertexCountsAttr()
        fvi_attr = mesh.CreateFaceVertexIndicesAttr()
        if conn is not None:
            fvc_attr.Set([3] * (conn.size // 3), 0)
            fvi_attr.Set(conn, 0)

        primvarsAPI = UsdGeom.PrimvarsAPI(mesh)
        texCoords = primvarsAPI.CreatePrimvar(
            "st", Sdf.ValueTypeNames.TexCoord2fArray, UsdGeom.Tokens.varying
        )
        texCoords.SetInterpolation("vertex")
        if tcoords is not None and variable is not None:
            texCoords.Set(tcoords, 0)

        stage.SetDefaultPrim(part_prim)
        stage.SetStartTimeCode(0)
        stage.SetEndTimeCode(0)

        self.create_dsg_material(
            stage,
            mesh,
            str(surfaces_prim.GetPath()),
            diffuse=diffuse,
            variable=variable,
            mat_info=mat_info,
        )

        stage.Save()
        return True

    def create_dsg_mesh_block(
        self,
        part: Part,
        name,
        id,
        part_hash,
        parent_prim,
        verts,
        conn,
        normals,
        tcoords,
        matrix=[1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0],
        diffuse=[1.0, 1.0, 1.0, 1.0],
        variable=None,
        timeline=[0.0, 0.0],
        first_timestep=False,
        mat_info={},
    ):
        if self._stage is None:
            return

        # 1D texture map for variables https://graphics.pixar.com/usd/release/tut_simple_shading.html
        # create the part usd object
        part_base_name = self.clean_name(name)
        partname = part_base_name + part_hash.hexdigest()
        stage_name = "/Parts/" + partname + self._ext
        part_stage_url = self.stage_url(os.path.join("Parts", partname + self._ext))

        # Make the manifest file - once for all timesteps
        part_manifest_url_relative = "./Parts/" + part_base_name + "_manifest" + self._ext
        part_manifest_url = self.stage_url(part_manifest_url_relative)
        created_file = self.create_dsg_surfaces_file(
            part_manifest_url,
            str(parent_prim.GetPath()),
            None,
            None,
            None,
            None,
            diffuse,
            variable,
            mat_info,
            True,
        )
        if created_file:
            self._stage.GetRootLayer().subLayerPaths.append(part_manifest_url_relative)

        # Make the per-timestep file
        created_file = self.create_dsg_surfaces_file(
            part_stage_url,
            str(parent_prim.GetPath()),
            verts,
            normals,
            conn,
            tcoords,
            diffuse,
            variable,
            mat_info,
            False,
        )

        # Glue the file into the main stage
        path = parent_prim.GetPath().AppendChild("surfaces")
        surfaces_prim = self._stage.OverridePrim(path)
        self.add_timestep_valueclip(
            part_base_name,
            "surfaces",
            surfaces_prim,
            part_manifest_url_relative,
            timeline,
            stage_name,
        )

        return part_stage_url

    def get_time_files(self, part_name: str, mesh_type: str):
        if part_name not in self._time_files:
            self._time_files[part_name] = {"surfaces": [], "lines": [], "points": []}
        return self._time_files[part_name][mesh_type]

    def add_timestep_valueclip(
        self,
        part_name: str,
        mesh_type: str,
        part_prim: UsdGeom.Xform,
        manifest_path: str,
        timeline: List[float],
        stage_name: str,
    ) -> None:
        clips_api = Usd.ClipsAPI(part_prim)
        asset_path = "." + stage_name

        time_files = self.get_time_files(part_name, mesh_type)

        if len(time_files) == 0 or time_files[-1][0] != asset_path:
            time_files.append((asset_path, timeline[0]))
            clips_api.SetClipAssetPaths([time_file[0] for time_file in time_files])
            clips_api.SetClipActive(
                [
                    (time_file[1] * self._time_codes_per_second, ii)
                    for ii, time_file in enumerate(time_files)
                ]
            )
            clips_api.SetClipTimes(
                [
                    (time_file[1] * self._time_codes_per_second, 0)
                    for ii, time_file in enumerate(time_files)
                ]
            )
            clips_api.SetClipPrimPath(str(part_prim.GetPath()))
            clips_api.SetClipManifestAssetPath(Sdf.AssetPath(manifest_path))

    # Common code to create the part manifest file and the file per timestep
    def create_dsg_lines_file(
        self,
        file_url,  # SdfPath, location on disk
        part_path: str,  # base path name, such as "/Root/Case_1/Isosurface_part"
        verts,
        width: float,
        tcoords,
        diffuse,
        var_cmd,
        mat_info,
        is_manifest: bool,
    ):
        if is_manifest and file_url in self._old_stages:
            return False
        if not is_manifest and os.path.exists(file_url):
            return False

        stage = Usd.Stage.CreateNew(file_url)
        UsdGeom.SetStageUpAxis(stage, self._up_axis)
        UsdGeom.SetStageMetersPerUnit(stage, 1.0 / self._units_per_meter)
        self._old_stages.append(file_url)

        part_prim = stage.OverridePrim(part_path)

        lines_prim = self.create_xform_node(stage, part_path + "/lines")

        lines = UsdGeom.BasisCurves.Define(stage, str(lines_prim.GetPath()) + "/Lines")
        lines.CreateDoubleSidedAttr().Set(True)
        pt_attr = lines.CreatePointsAttr()
        vc_attr = lines.CreateCurveVertexCountsAttr()
        if verts is not None:
            pt_attr.Set(verts, 0)
            vc_attr.Set([2] * (verts.size // 6), 0)
        lines.CreatePurposeAttr().Set("render")
        lines.CreateTypeAttr().Set("linear")
        lines.CreateWidthsAttr([width])
        lines.SetWidthsInterpolation("constant")

        # Rounded endpoint are a primvar
        primvarsAPI = UsdGeom.PrimvarsAPI(lines)
        endCaps = primvarsAPI.CreatePrimvar(
            "endcaps", Sdf.ValueTypeNames.Int, UsdGeom.Tokens.constant
        )
        endCaps.Set(2)  # Rounded = 2

        prim = lines.GetPrim()
        wireframe = width == 0.0
        prim.CreateAttribute("omni:scene:visualization:drawWireframe", Sdf.ValueTypeNames.Bool).Set(
            wireframe
        )

        if (tcoords is not None) and var_cmd:
            primvarsAPI = UsdGeom.PrimvarsAPI(lines)
            texCoords = primvarsAPI.CreatePrimvar(
                "st", Sdf.ValueTypeNames.TexCoord2fArray, UsdGeom.Tokens.varying
            )
            if tcoords is not None and var_cmd is not None:
                texCoords.Set(tcoords, 0)
            texCoords.SetInterpolation("vertex")
        stage.SetDefaultPrim(part_prim)
        stage.SetStartTimeCode(0)
        stage.SetEndTimeCode(0)

        self.create_dsg_material(
            stage,
            lines,
            str(lines_prim.GetPath()),
            diffuse=diffuse,
            variable=var_cmd,
            mat_info=mat_info,
        )
        stage.Save()
        return True

    def create_dsg_lines(
        self,
        name,
        id,
        part_hash,
        parent_prim,
        verts,
        tcoords,
        width,
        matrix=[1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0],
        diffuse=[1.0, 1.0, 1.0, 1.0],
        variable=None,
        timeline=[0.0, 0.0],
        first_timestep=False,
        mat_info={},
    ):
        # include the line width in the hash
        part_hash.update(str(self.line_width).encode("utf-8"))

        part_base_name = self.clean_name(name) + "_l"
        partname = part_base_name + part_hash.hexdigest()
        stage_name = "/Parts/" + partname + self._ext
        part_stage_url = self.stage_url(os.path.join("Parts", partname + self._ext))

        # Make the manifest file - once for all timesteps
        part_manifest_url_relative = "./Parts/" + part_base_name + "_manifest" + self._ext
        part_manifest_url = self.stage_url(part_manifest_url_relative)
        created_file = self.create_dsg_lines_file(
            part_manifest_url,
            str(parent_prim.GetPath()),
            None,
            width,
            None,
            diffuse,
            variable,
            mat_info,
            True,
        )
        if created_file:
            self._stage.GetRootLayer().subLayerPaths.append(part_manifest_url_relative)

        # Make the per-timestep file
        created_file = self.create_dsg_lines_file(
            part_stage_url,
            str(parent_prim.GetPath()),
            verts,
            width,
            tcoords,
            diffuse,
            variable,
            mat_info,
            False,
        )

        # Glue the file into the main stage
        path = parent_prim.GetPath().AppendChild("lines")
        lines_prim = self._stage.OverridePrim(path)
        self.add_timestep_valueclip(
            part_base_name, "lines", lines_prim, part_manifest_url_relative, timeline, stage_name
        )

        return part_stage_url

    # Common code to create the part manifest file and the file per timestep
    def create_dsg_points_file(
        self,
        file_url,  # SdfPath, location on disk
        part_path: str,  # base path name, such as "/Root/Case_1/Isosurface_part"
        verts,
        sizes,
        colors,
        default_size: float,
        default_color,
        is_manifest: bool,
    ):
        if is_manifest and file_url in self._old_stages:
            return False
        if not is_manifest and os.path.exists(file_url):
            return False

        stage = Usd.Stage.CreateNew(file_url)
        UsdGeom.SetStageUpAxis(stage, self._up_axis)
        UsdGeom.SetStageMetersPerUnit(stage, 1.0 / self._units_per_meter)
        self._old_stages.append(file_url)

        part_prim = stage.OverridePrim(part_path)
        points = UsdGeom.Points.Define(stage, part_path + "/points")
        pt_attr = points.CreatePointsAttr()
        w_attr = points.CreateWidthsAttr()
        if verts is not None:
            pt_attr.Set(verts, 0)
            if sizes is not None and sizes.size == (verts.size // 3):
                w_attr.Set(sizes, 0)
            else:
                w_attr.Set([default_size] * (verts.size // 3), 0)

        colorAttr = points.GetPrim().GetAttribute("primvars:displayColor")
        colorAttr.SetMetadata("interpolation", "vertex")
        if verts is not None:
            if colors is not None and colors.size == verts.size:
                colorAttr.Set(colors, 0)
            else:
                colorAttr.Set([default_color[0:3]] * (verts.size // 3), 0)

        stage.SetDefaultPrim(part_prim)
        stage.SetStartTimeCode(0)
        stage.SetEndTimeCode(0)

        stage.Save()
        return True

    def create_dsg_points(
        self,
        name,
        id,
        part_hash,
        parent_prim,
        verts,
        sizes,
        colors,
        matrix=[1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0, 0.0, 0.0, 0.0, 1.0],
        default_size=1.0,
        default_color=[1.0, 1.0, 1.0, 1.0],
        timeline=[0.0, 0.0],
        first_timestep=False,
    ):
        part_base_name = self.clean_name(name) + "_p"
        partname = part_base_name + part_hash.hexdigest()
        stage_name = "/Parts/" + partname + self._ext
        part_stage_url = self.stage_url(os.path.join("Parts", partname + self._ext))

        # Make the manifest file - once for all timesteps
        part_manifest_url_relative = "./Parts/" + part_base_name + "_manifest" + self._ext
        part_manifest_url = self.stage_url(part_manifest_url_relative)
        created_file = self.create_dsg_points_file(
            part_manifest_url,
            str(parent_prim.GetPath()),
            None,
            None,
            None,
            default_size,
            default_color,
            True,
        )
        if created_file:
            self._stage.GetRootLayer().subLayerPaths.append(part_manifest_url_relative)

        # Make the per-timestep file
        created_file = self.create_dsg_points_file(
            part_stage_url,
            str(parent_prim.GetPath()),
            verts,
            sizes,
            colors,
            default_size,
            default_color,
            False,
        )

        # Glue the file into the main stage
        path = parent_prim.GetPath().AppendChild("points")
        points_prim = self._stage.OverridePrim(path)
        self.add_timestep_valueclip(
            part_base_name, "points", points_prim, part_manifest_url_relative, timeline, stage_name
        )

        return part_stage_url

    def create_dsg_material(
        self, stage, mesh, root_name, diffuse=[1.0, 1.0, 1.0, 1.0], variable=None, mat_info={}
    ):
        # https://graphics.pixar.com/usd/release/spec_usdpreviewsurface.html
        # Use ior==1.0 to be more like EnSight - rays of light do not bend when passing through transparent objs
        material = UsdShade.Material.Define(stage, root_name + "/Material")
        pbrShader = UsdShade.Shader.Define(stage, root_name + "/Material/PBRShader")
        pbrShader.CreateIdAttr("UsdPreviewSurface")
        smoothness = mat_info.get("smoothness", 0.0)
        pbrShader.CreateInput("roughness", Sdf.ValueTypeNames.Float).Set(1.0 - smoothness)
        metallic = mat_info.get("metallic", 0.0)
        pbrShader.CreateInput("metallic", Sdf.ValueTypeNames.Float).Set(metallic)
        opacity = mat_info.get("opacity", diffuse[3])
        pbrShader.CreateInput("opacity", Sdf.ValueTypeNames.Float).Set(opacity)
        pbrShader.CreateInput("ior", Sdf.ValueTypeNames.Float).Set(1.0)
        pbrShader.CreateInput("useSpecularWorkflow", Sdf.ValueTypeNames.Int).Set(1)
        if variable:
            stReader = UsdShade.Shader.Define(stage, root_name + "/Material/stReader")
            stReader.CreateIdAttr("UsdPrimvarReader_float2")
            diffuseTextureSampler = UsdShade.Shader.Define(
                stage, root_name + "/Material/diffuseTexture"
            )
            diffuseTextureSampler.CreateIdAttr("UsdUVTexture")
            name = self.clean_name(variable.name)
            filename = f"./Textures/palette_{name}.png"
            diffuseTextureSampler.CreateInput("file", Sdf.ValueTypeNames.Asset).Set(filename)
            diffuseTextureSampler.CreateInput("st", Sdf.ValueTypeNames.Float2).ConnectToSource(
                stReader.ConnectableAPI(), "result"
            )
            diffuseTextureSampler.CreateOutput("rgb", Sdf.ValueTypeNames.Float3)
            pbrShader.CreateInput("diffuseColor", Sdf.ValueTypeNames.Color3f).ConnectToSource(
                diffuseTextureSampler.ConnectableAPI(), "rgb"
            )
            stInput = material.CreateInput("frame:stPrimvarName", Sdf.ValueTypeNames.Token)
            stInput.Set("st")
            stReader.CreateInput("varname", Sdf.ValueTypeNames.Token).ConnectToSource(stInput)
        else:
            # The colors are a mixture of content from the DSG PART protocol buffer
            # and the JSON material block from the material_name field.
            kd = 1.0
            diffuse_color = [diffuse[0], diffuse[1], diffuse[2]]
            ke = 1.0
            emissive_color = [0.0, 0.0, 0.0]
            ks = 1.0
            specular_color = [0.0, 0.0, 0.0]
            mat_name = mat_info.get("name", "")
            if mat_name.startswith("ensight"):
                diffuse_color = mat_info.get("diffuse", diffuse_color)
                if mat_name != "ensight/Default":
                    ke = mat_info.get("ke", ke)
                    emissive_color = mat_info.get("emissive", emissive_color)
                    ks = mat_info.get("ks", ks)
                    specular_color = mat_info.get("specular", specular_color)
            # Set the colors
            color = Gf.Vec3f(diffuse_color[0] * kd, diffuse_color[1] * kd, diffuse_color[2] * kd)
            pbrShader.CreateInput("diffuseColor", Sdf.ValueTypeNames.Color3f).Set(color)
            color = Gf.Vec3f(emissive_color[0] * ke, emissive_color[1] * ke, emissive_color[2] * ke)
            pbrShader.CreateInput("emissiveColor", Sdf.ValueTypeNames.Color3f).Set(color)
            color = Gf.Vec3f(specular_color[0] * ks, specular_color[1] * ks, specular_color[2] * ks)
            pbrShader.CreateInput("specularColor", Sdf.ValueTypeNames.Color3f).Set(color)

        material.CreateSurfaceOutput().ConnectToSource(pbrShader.ConnectableAPI(), "surface")
        mat_binding_api = UsdShade.MaterialBindingAPI.Apply(mesh.GetPrim())
        mat_binding_api.Bind(material)

        return material

    def create_dsg_variable_textures(self, variables):
        with tempfile.TemporaryDirectory() as tempdir:
            # make folder:   {tempdir}/scratch/Textures/{palette_*.png}
            os.makedirs(f"{tempdir}/scratch/Textures", exist_ok=True)
            for var in variables.values():
                data = bytearray(var.texture)
                n_pixels = int(len(data) / 4)
                row = []
                for i in range(n_pixels):
                    row.append(data[i * 4 + 0])
                    row.append(data[i * 4 + 1])
                    row.append(data[i * 4 + 2])
                io = png.Writer(width=n_pixels, height=2, bitdepth=8, greyscale=False)
                rows = [row, row]
                name = self.clean_name(var.name)
                with open(f"{tempdir}/scratch/Textures/palette_{name}.png", "wb") as fp:
                    io.write(fp, rows)
            uriPath = self._destinationPath + "/Parts/Textures"
            shutil.rmtree(uriPath, ignore_errors=True, onerror=None)
            shutil.copytree(f"{tempdir}/scratch/Textures", uriPath)

    def create_xform_node(self, stage, name):
        xform_node = UsdGeom.Xform.Define(stage, name)
        xform_node.AddTranslateOp().Set((0, 0, 0))
        xform_node.AddRotateXYZOp().Set((0, 0, 0))
        xform_node.AddScaleOp().Set((1, 1, 1))
        if self._centroid is not None:
            xform_api = UsdGeom.XformCommonAPI(xform_node.GetPrim())
            xform_api.SetPivot(Gf.Vec3f(self._centroid) * self._units_per_meter)
        return xform_node

    def create_dsg_root(self):
        root_name = "/Root"
        self.create_xform_node(self._stage, root_name)

        # Define the defaultPrim as the /Root prim
        root_prim = self._stage.GetPrimAtPath(root_name)
        self._stage.SetDefaultPrim(root_prim)
        return root_prim

    def update_camera(self, camera):
        if camera is not None:
            cam_name = "/Root/Cam"
            cam_prim = UsdGeom.Xform.Define(self._stage, cam_name)
            s = self._units_per_meter
            cam_pos = Gf.Vec3d(camera.lookfrom[0], camera.lookfrom[1], camera.lookfrom[2]) * s
            target_pos = Gf.Vec3d(camera.lookat[0], camera.lookat[1], camera.lookat[2]) * s

            up_vec = Gf.Vec3d(camera.upvector[0], camera.upvector[1], camera.upvector[2])
            cam_prim = self._stage.GetPrimAtPath(cam_name)
            geom_cam = UsdGeom.Camera(cam_prim)
            if not geom_cam:
                geom_cam = UsdGeom.Camera.Define(self._stage, cam_name)
            # Set camera values
            # center of interest attribute unique for Kit defines the pivot for tumbling the camera
            # Set as an attribute on the prim
            coi_attr = cam_prim.GetAttribute("omni:kit:centerOfInterest")
            if not coi_attr.IsValid():
                coi_attr = cam_prim.CreateAttribute(
                    "omni:kit:centerOfInterest", Sdf.ValueTypeNames.Vector3d
                )
            coi_attr.Set(target_pos)
            # get the camera
            cam = geom_cam.GetCamera()
            # LOL, not sure why is might be correct, but so far it seems to work???
            cam.focalLength = camera.fieldofview
            dist = (target_pos - cam_pos).GetLength()
            cam.clippingRange = Gf.Range1f(0.1 * dist, 1000.0 * dist)
            look_at = Gf.Matrix4d()
            look_at.SetLookAt(cam_pos, target_pos, up_vec)
            trans_row = look_at.GetRow(3)
            trans_row = Gf.Vec4d(-trans_row[0], -trans_row[1], -trans_row[2], trans_row[3])
            look_at.SetRow(3, trans_row)
            cam.transform = look_at

            # set the updated camera
            geom_cam.SetFromCamera(cam)

    def create_dsg_group(
        self,
        name: str,
        parent_prim,
        obj_type: Any = None,
        matrix: List[float] = [
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
            0.0,
            0.0,
            0.0,
            0.0,
            1.0,
        ],
    ):
        path = parent_prim.GetPath().AppendChild(self.clean_name(name))
        group_prim = UsdGeom.Xform.Get(self._stage, path)
        if not group_prim:
            group_prim = self.create_xform_node(self._stage, path)

            # Map kinds
            kind = Kind.Tokens.group
            if obj_type == "ENS_CASE":
                kind = Kind.Tokens.assembly
            elif obj_type == "ENS_PART":
                kind = Kind.Tokens.component
            Usd.ModelAPI(group_prim).SetKind(kind)
            group_prim.GetPrim().SetDisplayName(name)
            logging.info(f"Created group:'{name}' {str(obj_type)}")

        return group_prim

    def uploadMaterial(self):
        uriPath = self._destinationPath + "/Materials"
        shutil.rmtree(uriPath, ignore_errors=True, onerror=None)
        fullpath = os.path.join(os.path.dirname(__file__), "resources", "Materials")
        shutil.copytree(fullpath, uriPath)

    # Create a dome light in the scene.
    def createDomeLight(self, texturePath):
        newLight = UsdLux.DomeLight.Define(self._stage, "/Root/DomeLight")
        newLight.CreateIntensityAttr(2200.0)
        newLight.CreateTextureFileAttr(texturePath)
        newLight.CreateTextureFormatAttr("latlong")

        # Set rotation on domelight
        xForm = newLight
        rotateOp = xForm.AddXformOp(UsdGeom.XformOp.TypeRotateZYX, UsdGeom.XformOp.PrecisionFloat)
        rotateOp.Set(Gf.Vec3f(270, 0, 0))


class OmniverseUpdateHandler(UpdateHandler):
    """
    Implement the Omniverse glue to a DSGSession instance
    """

    def __init__(self, omni: OmniverseWrapper):
        super().__init__()
        self._omni = omni
        self._group_prims: Dict[int, Any] = dict()
        self._root_prim = None
        self._sent_textures = False
        self._case_xform_applied_to_camera = False

    def add_group(self, id: int, view: bool = False) -> None:
        super().add_group(id, view)
        group = self.session.groups[id]

        if not view:
            # Capture changes in line/sphere sizes if it was not set from cli
            width = self.get_dsg_cmd_attribute(group, "ANSYS_linewidth")
            if width:
                try:
                    self._omni.line_width = float(width)
                except ValueError:
                    pass

            parent_prim = self._group_prims[group.parent_id]
            # get the EnSight object type and the transform matrix
            obj_type = self.get_dsg_cmd_attribute(group, "ENS_OBJ_TYPE")
            matrix = group.matrix4x4
            # Is this a "case" group (it will contain part of the camera view in the matrix)
            if obj_type == "ENS_CASE":
                if self.session.scene_bounds is not None:
                    midx = (self.session.scene_bounds[3] + self.session.scene_bounds[0]) * 0.5
                    midy = (self.session.scene_bounds[4] + self.session.scene_bounds[1]) * 0.5
                    midz = (self.session.scene_bounds[5] + self.session.scene_bounds[2]) * 0.5
                    self._omni._centroid = [midx, midy, midz]

                if not self.session.vrmode and not self._case_xform_applied_to_camera:
                    # if in camera mode, we need to update the camera matrix so we can
                    # use the identity matrix on this group.  The camera should have been
                    # created in the "view" handler
                    self._case_xform_applied_to_camera = True
                    cam_name = "/Root/Cam"
                    cam_prim = self._omni._stage.GetPrimAtPath(cam_name)  # type: ignore
                    geom_cam = UsdGeom.Camera(cam_prim)
                    # get the camera
                    cam = geom_cam.GetCamera()
                    c = cam.transform
                    m = Gf.Matrix4d(*matrix).GetTranspose()
                    s = self._omni._units_per_meter
                    trans = m.GetRow(3)
                    trans = Gf.Vec4d(trans[0] * s, trans[1] * s, trans[2] * s, trans[3])
                    m.SetRow(3, trans)
                    # move the model transform to the camera transform
                    cam.transform = c * m.GetInverse()

                    # Determine if the camera is principally more Y, or Z up.  X up not supported.
                    # Omniverse' built in navigator tries to keep this direction up
                    # If the view is principally -Y, there is no good choice.  +Y is least bad.
                    cam_upvec = Gf.Vec4d(0, 1, 0, 0) * cam.transform
                    if abs(cam_upvec[1]) >= abs(cam_upvec[2]):
                        self._up_axis = UsdGeom.Tokens.y
                    else:
                        self._up_axis = UsdGeom.Tokens.z
                    UsdGeom.SetStageUpAxis(self._omni._stage, self._up_axis)

                    # set the updated camera
                    geom_cam.SetFromCamera(cam)
                    # apply the inverse cam transform to move the center of interest
                    # from data space to camera space
                    coi_attr = cam_prim.GetAttribute("omni:kit:centerOfInterest")
                    if coi_attr.IsValid():
                        coi_data = coi_attr.Get()
                        coi_cam = (
                            Gf.Vec4d(coi_data[0], coi_data[1], coi_data[2], 1.0)
                            * cam.transform.GetInverse()
                        )
                        coi_attr.Set(
                            Gf.Vec3d(
                                0,
                                0,
                                coi_cam[2] / coi_cam[3],
                            )
                        )
                    # use the camera view by default
                    self._omni._stage.GetRootLayer().customLayerData = {  # type: ignore
                        "cameraSettings": {"boundCamera": "/Root/Cam"}
                    }
                matrix = [
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                    0.0,
                    0.0,
                    0.0,
                    0.0,
                    1.0,
                ]
            prim = self._omni.create_dsg_group(
                group.name, parent_prim, matrix=matrix, obj_type=obj_type
            )
            self._group_prims[id] = prim
        else:
            # Map a view command into a new Omniverse stage and populate it with materials/lights.
            # Create a new root stage in Omniverse

            # Create or update the root group/camera
            if not self.session.vrmode:
                self._omni.update_camera(camera=group)

            # record
            self._group_prims[id] = self._root_prim

            if self._omni._stage is not None:
                self._omni._stage.SetStartTimeCode(
                    self.session.time_limits[0] * self._omni._time_codes_per_second
                )
                self._omni._stage.SetEndTimeCode(
                    self.session.time_limits[1] * self._omni._time_codes_per_second
                )
                self._omni._stage.SetTimeCodesPerSecond(self._omni._time_codes_per_second)

            # Send the variable textures.  Safe to do so once the first view is processed.
            if not self._sent_textures:
                self._omni.create_dsg_variable_textures(self.session.variables)
                self._sent_textures = True

    def add_variable(self, id: int) -> None:
        super().add_variable(id)

    def finalize_part(self, part: Part) -> None:
        # generate an Omniverse compliant mesh from the Part
        if part is None or part.cmd is None:
            return
        parent_prim = self._group_prims[part.cmd.parent_id]
        obj_id = self.session.mesh_block_count
        matrix = part.cmd.matrix4x4
        name = part.cmd.name
        color = [
            part.cmd.fill_color[0] * part.cmd.diffuse,
            part.cmd.fill_color[1] * part.cmd.diffuse,
            part.cmd.fill_color[2] * part.cmd.diffuse,
            part.cmd.fill_color[3],
        ]

        mat_info = part.material()
        if part.cmd.render == part.cmd.CONNECTIVITY:
            has_triangles = False
            command, verts, conn, normals, tcoords, var_cmd = part.nodal_surface_rep()
            if verts is not None:
                verts = numpy.multiply(verts, self._omni._units_per_meter)
            if command is not None:
                has_triangles = True
                # Generate the mesh block
                _ = self._omni.create_dsg_mesh_block(
                    part,
                    name,
                    obj_id,
                    part.hash,
                    parent_prim,
                    verts,
                    conn,
                    normals,
                    tcoords,
                    matrix=matrix,
                    diffuse=color,
                    variable=var_cmd,
                    timeline=self.session.cur_timeline,
                    first_timestep=(self.session.cur_timeline[0] == self.session.time_limits[0]),
                    mat_info=mat_info,
                )
            command, verts, tcoords, var_cmd = part.line_rep()
            if verts is not None:
                verts = numpy.multiply(verts, self._omni._units_per_meter)
            if command is not None:
                # If there are no triangle (ideally if these are not hidden line
                # edges), then use the base color for the part.  If there are
                # triangles, then assume these are hidden line edges and use the
                # line_color.
                line_color = color
                if has_triangles:
                    line_color = [
                        part.cmd.line_color[0] * part.cmd.diffuse,
                        part.cmd.line_color[1] * part.cmd.diffuse,
                        part.cmd.line_color[2] * part.cmd.diffuse,
                        part.cmd.line_color[3],
                    ]
                # TODO: texture coordinates on lines are currently invalid in Omniverse
                var_cmd = None
                tcoords = None
                # line info can come from self or our parent group
                width = self._omni.line_width
                # Allow the group to override
                group = self.session.find_group_pb(part.cmd.parent_id)
                if group:
                    try:
                        width = float(group.attributes.get("ANSYS_linewidth", str(width)))
                    except ValueError:
                        pass
                if width < 0.0:
                    tmp = verts.reshape(-1, 3)
                    mins = numpy.min(tmp, axis=0)
                    maxs = numpy.max(tmp, axis=0)
                    dx = maxs[0] - mins[0]
                    dy = maxs[1] - mins[1]
                    dz = maxs[2] - mins[2]
                    diagonal = math.sqrt(dx * dx + dy * dy + dz * dz)
                    width = diagonal * math.fabs(width) / self._omni._units_per_meter
                    if self._omni.line_width < 0.0:
                        self._omni.line_width = width
                width = width * self._omni._units_per_meter
                # Generate the lines
                _ = self._omni.create_dsg_lines(
                    name,
                    obj_id,
                    part.hash,
                    parent_prim,
                    verts,
                    tcoords,
                    width,
                    matrix=matrix,
                    diffuse=line_color,
                    variable=var_cmd,
                    timeline=self.session.cur_timeline,
                    first_timestep=(self.session.cur_timeline[0] == self.session.time_limits[0]),
                )

        elif part.cmd.render == part.cmd.NODES:
            command, verts, sizes, colors, var_cmd = part.point_rep()
            if verts is not None:
                verts = numpy.multiply(verts, self._omni._units_per_meter)
            if sizes is not None:
                sizes = numpy.multiply(sizes, self._omni._units_per_meter)
            if command is not None:
                _ = self._omni.create_dsg_points(
                    name,
                    obj_id,
                    part.hash,
                    parent_prim,
                    verts,
                    sizes,
                    colors,
                    matrix=matrix,
                    default_size=part.cmd.node_size_default * self._omni._units_per_meter,
                    default_color=color,
                    timeline=self.session.cur_timeline,
                    first_timestep=(self.session.cur_timeline[0] == self.session.time_limits[0]),
                )
        super().finalize_part(part)

    def start_connection(self) -> None:
        super().start_connection()

    def end_connection(self) -> None:
        super().end_connection()

    def begin_update(self) -> None:
        super().begin_update()
        # restart the name tables
        self._omni.clear_cleaned_names()
        # clear the group Omni prims list
        self._group_prims = dict()
        self._case_xform_applied_to_camera = False

        self._omni.create_new_stage()
        self._root_prim = self._omni.create_dsg_root()
        # Create a distance and dome light in the scene
        self._omni.createDomeLight("./Materials/000_sky.exr")
        # Upload a material to the Omniverse server
        self._omni.uploadMaterial()
        self._sent_textures = False

    def end_update(self) -> None:
        super().end_update()
        # Stage update complete
        self._omni.save_stage()



from typing import TYPE_CHECKING, Any, ContextManager, Union

if TYPE_CHECKING:
    try:
        import ensight
    except ImportError:
        from ansys.api.pyensight import ensight_api


class Support:
    """Provides the ``ensight.utils.support`` interface.

    This class provides a collection of general utility functions and objects
    that can be used to simplify various issues that come up when using EnSight
    and PyEnSight.

    This class is instantiated as ``ensight.utils.support`` in EnSight Python
    and as ``Session.ensight.utils.support`` in PyEnSight. The constructor is
    passed the interface, which serves as the ``ensight`` module for either
    case. As a result, the methods can be accessed as ``ensight.utils.support.*``
    in EnSight Python or ``session.ensight.utils.support.*`` in PyEnSight.

    Parameters
    ----------
    interface :
        Entity that provides the ``ensight`` namespace. In the case of
        EnSight Python, the ``ensight`` module is passed. In the case
        of PyEnSight, ``Session.ensight`` is passed.
    """

    def __init__(self, interface: Union["ensight_api.ensight", "ensight"]):
        self._ensight = interface

    def scoped_name(self, obj: Any, native_exceptions: bool = False) -> ContextManager:
        """Allow for the use of ``with`` to shorten APIs.

        In the EnSight and PyEnsight APIs, the interfaces can become lengthy.
        This class makes it possible to shorten APIs for modules, classes,
        and namespaces.  The native_exceptions keyword can be used to enable
        exceptions for EnSight native Python API.  By default, an invalid
        native API call like ``ensight.part.select_begin(-9999)`` will return
        -1.  If native_exceptions is True, a Python exception will be thrown.
        The scope of this operational change parallels the scoped_name()
        instance.

        Parameters
        ----------
        obj: Any
            The object for which to generate a simplified namespace.
        native_exceptions: bool
            If True, then EnSight native Python API exceptions are enabled.
            The default is False.

        Returns
        -------
        The passed object wrapped in a context manager that can be used as a
        simplified namespace.

        Examples
        --------
        >>> sn = s.ensight.utils.support.scoped_name
        >>> with sn(s.ensight.objs.core) as core, sn(s.ensight.objs.enums) as enums:
        >>>     print(core.PARTS.find(True, enums.VISIBLE))

        >>> sn = ensight.utils.support.scoped_name
        >>> with sn(ensight.objs.core) as core, sn(ensight.objs.enums) as enums:
        >>>     print(core.PARTS.find(True, enums.VISIBLE))

        >>> sn = ensight.utils.support.scoped_name
        >>> with sn(ensight.part, native_exceptions=True) as part:
        >>>     part.select_begin(-9999)
        """
        return ScopedName(self._ensight, obj, native_exceptions=native_exceptions)


class ScopedName:
    """Allow for the use of ``with`` to shorten APIs.

    In the EnSight and PyEnsight APIs, the interfaces can become lengthy.
    This class makes it possible to shorten APIs for modules, classes,
    and namespaces.
    """

    def __init__(
        self,
        interface: Union["ensight_api.ensight", "ensight"],
        obj: Any,
        native_exceptions: bool = False,
    ):
        self._obj = obj
        self._ensight = interface
        self._old_raise = None
        if native_exceptions:
            # if we are being asked to enable exceptions, record what to restore it to
            self._old_raise = self._ensight.query("SENDMESG_RAISE")

    def __enter__(self) -> Any:
        if self._old_raise is not None:
            # if a restore value is set, enable them
            self._ensight.sendmesgoptions(exception=1)
        return self._obj

    def __exit__(self, exc_type, exc_value, exc_trace):
        if self._old_raise is not None:
            # if the restore value is set, restore it here
            self._ensight.sendmesgoptions(exception=self._old_raise)



"""Docker Launcher module.

The Docker Launcher module provides PyEnSight with the ability to launch a
:class:`Session<ansys.pyensight.core.Session>` instance using a local Docker
installation or a `PIM <https://github.com/ansys/pypim`_-launched Docker installation.

Examples:
    ::

        from ansys.pyensight.core import DockerLauncher
        launcher = DockerLauncher(data_directory="D:\\data")
        launcher.pull()
        session = launcher.start()
        session.close()

"""
import logging
import os
import re
import subprocess
from typing import TYPE_CHECKING, Any, Dict, Optional
import uuid
import warnings

try:
    import grpc
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError("The grpc module must be installed for DockerLauncher")
except Exception:  # pragma: no cover
    raise RuntimeError("Cannot initialize grpc")

try:
    import docker
except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError("The docker module must be installed for DockerLauncher")
except Exception:  # pragma: no cover
    raise RuntimeError("Cannot initialize Docker")

from ansys.pyensight.core.common import (
    GRPC_WARNING_MESSAGE,
    find_unused_ports,
    get_file_service,
    grpc_version_check,
    launch_enshell_interface,
    populate_service_host_port,
    pull_image,
)
from ansys.pyensight.core.launcher import Launcher
import ansys.pyensight.core.session
from ansys.pyensight.core.session import Session

if TYPE_CHECKING:
    from docker import DockerClient
    from docker.models.containers import Container


class DockerLauncher(Launcher):
    """Creates a ``Session`` instance using a copy of EnSight hosted in a Docker container.

    This class allows you to either attach to a `PIM <https://github.com/ansys/pypim>`_-launched
    Docker container hosting EnSight or launch a local Docker container hosting EnSight.
    gRPC connections to EnShell and EnSight are established. A PyEnSight ``Session`` instance
    is returned upon successful launch and connection.

    Parameters
    ----------
    data_directory : str, optional
        Host directory to make into the container at ``/data``.
        The default is ``None``.
    docker_image_name : str, optional
        Name of the Docker image to use. The default is ``None``.
    use_dev : bool, optional
        Whether to use the latest ``ensight_dev`` Docker image. The
        default is ``False``. However, this value is overridden if the
        name of a Docker image is specified for the ``docker_image_name``
        parameter.
    channel :
        Existing gRPC channel to a running ``EnShell`` instance provided by PIM.
    pim_instance :
        PyPIM instance if using PIM (internal). The default is ``None``.
    timeout : float, optional
        Number of seconds to try a gRPC connection before giving up.
        This parameter is defined on the parent ``Launcher`` class,
        where the default is ``120``.
    use_egl : bool, optional
        Whether to use EGL hardware for accelerated graphics. The platform
        must be able to support this hardware. This parameter is defined on
        the parent ``Launcher`` class, where the default is ``False``.
    use_sos : int, optional
        Number of EnSight servers to use for SOS (Server of Server) mode.
        This parameter is defined on the parent ``Launcher`` class, where
        the default is ``None``, in which case SOS mode is not used.
    additional_command_line_options: list, optional
        Additional command line options to be used to launch EnSight.
        Arguments that contain spaces are not supported.
    grpc_use_tcp_sockets :
        If using gRPC, and if True, then allow TCP Socket based connections
        instead of only local connections.
    grpc_allow_network_connections :
        If using gRPC and using TCP Socket based connections, listen on all networks.
    grpc_disable_tls :
        If using gRPC and using TCP Socket based connections, disable TLS.
    grpc_uds_pathname :
        If using gRPC and using Unix Domain Socket based connections, explicitly
        set the pathname to the shared UDS file instead of using the default.

    Examples
    --------
    >>> from ansys.pyensight.core import DockerLauncher
    >>> launcher = DockerLauncher(data_directory="D:\\data")
    >>> launcher.pull()
    >>> session = launcher.start()
    >>> session.close()

    WARNING:
    Overriding the default values for these options: grpc_use_tcp_sockets, grpc_allow_network_connections,
    and grpc_disable_tls
    can possibly permit control of this computer and any data which resides on it.
    Modification of this configuration is not recommended.  Please see the
    documentation for your installed product for additional information.
    """

    def __init__(
        self,
        data_directory: Optional[str] = None,
        docker_image_name: Optional[str] = None,
        use_dev: bool = False,
        channel: Optional[grpc.Channel] = None,
        pim_instance: Optional[Any] = None,
        grpc_use_tcp_sockets: Optional[bool] = False,
        grpc_allow_network_connections: Optional[bool] = False,
        grpc_disable_tls: Optional[bool] = False,
        grpc_uds_pathname: Optional[str] = None,
        **kwargs,
    ) -> None:
        """Initialize DockerLauncher."""
        super().__init__(**kwargs)

        self._data_directory = data_directory
        self._enshell_grpc_channel = channel
        self._grpc_use_tcp_sockets = grpc_use_tcp_sockets
        self._grpc_allow_network_connections = grpc_allow_network_connections
        self._grpc_disable_tls = grpc_disable_tls
        self._grpc_uds_pathname = grpc_uds_pathname
        self._service_uris: Dict[Any, str] = {}
        self._image_name: Optional[str] = None
        self._docker_client: Optional["DockerClient"] = None
        self._container: Optional["Container"] = None
        self._enshell: Optional[Any] = None
        self._pim_instance: Optional[Any] = pim_instance

        # EnSight session secret key
        self._secret_key: str = str(uuid.uuid1())
        # temporary directory
        # it's in the ephemeral container, so just use "ensight's"
        # home directory within the container
        self._session_directory: str = "/home/ensight"
        # the Ansys / EnSight version we found in the container
        # to be reassigned later
        self._ansys_version: Optional[str] = None
        # EnShell's log file
        self._enshell_log_file = None
        # pointer to the file service object if available
        self._pim_file_service = None

        logging.debug(
            f"DockerLauncher data_dir={self._data_directory}\n"
            + f"  image_name={self._image_name}\n"
            + f"  use_dev={use_dev}\n"
        )

        if self._enshell_grpc_channel and self._pim_instance:
            service_set = ["grpc_private", "http", "ws"]
            # if self._launch_webui:
            #    service_set.append("webui")
            if not set(service_set).issubset(self._pim_instance.services):
                raise RuntimeError(
                    "If channel is specified, the PIM instance must have a list of length 3 "
                    + "containing the appropriate service URIs. It does not."
                )
            # grab the URIs for the 3 required services passed in from PIM
            self._service_host_port = populate_service_host_port(
                self._pim_instance, {}, webui=self._launch_webui
            )
            # attach to the file service if available
            self._pim_file_service = get_file_service(self._pim_instance)
            # if using PIM, we have a query parameter to append to http requests
            if self._pim_instance is not None:
                d = {"instance_name": self._pim_instance.name}
                self._add_query_parameters(d)
            #
            return

        # EnShell gRPC port, EnSight gRPC port, HTTP port, WSS port
        # skip 1999 as that internal to the container is used to the container for the VNC connection
        num_ports = 4
        if self._launch_webui:
            num_ports = 5
        ports = find_unused_ports(num_ports, avoid=[1999])
        if ports is None:  # pragma: no cover
            raise RuntimeError(
                "Unable to allocate local ports for EnSight session"
            )  # pragma: no cover
        self._service_host_port = {}
        self._service_host_port["grpc"] = ("127.0.0.1", ports[0])
        self._service_host_port["grpc_private"] = ("127.0.0.1", ports[1])
        self._service_host_port["http"] = ("127.0.0.1", ports[2])
        self._service_host_port["ws"] = ("127.0.0.1", ports[3])
        if self._launch_webui:
            self._service_host_port["webui"] = ("127.0.0.1", ports[4])

        # get the optional user-specified image name
        # Note: the default name needs to change over time...  TODO
        self._image_name = "ghcr.io/ansys-internal/ensight"
        if use_dev:
            self._image_name = "ghcr.io/ansys-internal/ensight_dev"
        if docker_image_name:
            self._image_name = docker_image_name

        # Load up Docker from the user's environment
        self._docker_client = docker.from_env()

    def ansys_version(self) -> Optional[str]:
        """Get the Ansys version (three-digit string) found in the Docker container.

        Returns
        -------
        str
            Ansys version or ``None`` if not found or not started.

        """
        return self._ansys_version

    def pull(self) -> None:
        """Pull the Docker image.

        Raises
        ------
        RuntimeError
            If the Docker image couldn't be pulled.

        """
        pull_image(self._docker_client, self._image_name)

    def _get_container_env(self) -> Dict:
        # Create the environmental variables
        # Environment to pass into the container
        container_env = {
            "ENSIGHT_SECURITY_TOKEN": self._secret_key,
            "WEBSOCKETSERVER_SECURITY_TOKEN": self._secret_key,
            "ENSIGHT_SESSION_TEMPDIR": self._session_directory,
        }

        # If for some reason, the ENSIGHT_ANSYS_LAUNCH is set previously,
        # honor that value, otherwise set it to "pyensight".  This allows
        # for an environmental setup to set the value to something else
        # (e.g. their "app").
        if "ENSIGHT_ANSYS_LAUNCH" not in os.environ:
            container_env["ENSIGHT_ANSYS_LAUNCH"] = "container"
        else:
            container_env["ENSIGHT_ANSYS_LAUNCH"] = os.environ["ENSIGHT_ANSYS_LAUNCH"]

        if self._pim_instance is None:
            container_env["ANSYSLMD_LICENSE_FILE"] = os.environ["ANSYSLMD_LICENSE_FILE"]
            if "ENSIGHT_ANSYS_APIP_CONFIG" in os.environ:
                container_env["ENSIGHT_ANSYS_APIP_CONFIG"] = os.environ["ENSIGHT_ANSYS_APIP_CONFIG"]

        if self._launch_webui:
            container_env["SIMBA_WEBSERVER_TOKEN"] = self._secret_key
            container_env["FLUENT_WEBSERVER_TOKEN"] = self._secret_key

        return container_env

    def start(self) -> "Session":
        """Start EnShell by running a local Docker EnSight image.

        Once this method starts EnShell, it connects over gRPC to EnShell running
        in the Docker container. Once connected, EnShell launches a copy of EnSight
        and WSS in the container. It then creates and binds a
        :class:`Session<ansys.pyensight.core.Session>` instance to the created
        EnSight gRPC connection and returns this instance.

        Returns
        -------
        object
            PyEnSight ``Session`` object instance.

        Raises
        ------
        RuntimeError
            Variety of error conditions.

        """
        tmp_session = super().start()
        if tmp_session:
            return tmp_session

        # Launch the EnSight Docker container locally as a detached container
        # initially running EnShell over the first gRPC port. Then launch EnSight
        # and other apps.

        # get the environment to pass to the container
        container_env = self._get_container_env()

        # Ports to map between the host and the container
        # If we're here in the code, then we're not using PIM
        # and we're not really using URIs where the hostname
        # is anything other than 127.0.0.1, so, we only need
        # to grab the port numbers.
        grpc_port = self._service_host_port["grpc"][1]
        ports_to_map = {
            str(self._service_host_port["grpc"][1])
            + "/tcp": str(self._service_host_port["grpc"][1]),
            str(self._service_host_port["grpc_private"][1])
            + "/tcp": str(self._service_host_port["grpc_private"][1]),
            str(self._service_host_port["http"][1])
            + "/tcp": str(self._service_host_port["http"][1]),
            str(self._service_host_port["ws"][1]) + "/tcp": str(self._service_host_port["ws"][1]),
        }
        if self._launch_webui:
            ports_to_map.update(
                {
                    str(self._service_host_port["webui"][1])
                    + "/tcp": str(self._service_host_port["webui"][1])
                }
            )
        # The data directory to map into the container
        data_volume = None
        if self._data_directory:
            data_volume = {self._data_directory: {"bind": "/data", "mode": "rw"}}

        # Start the container in detached mode with EnShell as a
        # gRPC server as the command
        #
        enshell_cmd = "-app -v 3 -grpc_server " + str(grpc_port)
        if self._grpc_use_tcp_sockets:
            enshell_cmd += " -grpc_use_tcp_sockets"
        if self._grpc_allow_network_connections:
            enshell_cmd += " -grpc_allow_network_connections"
        if self._grpc_disable_tls:
            enshell_cmd += " -grpc_disable_tls"
        if self._grpc_uds_pathname:
            enshell_cmd += " -grpc_uds_pathname " + self._grpc_uds_pathname

        use_egl = self._use_egl()

        logging.debug("Starting Container...\n")
        if data_volume:
            if use_egl:  # pragma: no cover
                if self._docker_client:
                    self._container = self._docker_client.containers.run(  # pragma: no cover
                        self._image_name,
                        command=enshell_cmd,
                        volumes=data_volume,
                        environment=container_env,
                        device_requests=[
                            docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])
                        ],
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                        user="ensight:ensight",
                    )
            else:
                logging.debug(f"Running container {self._image_name} with cmd {enshell_cmd}\n")
                logging.debug(f"ports to map: {ports_to_map}\n")
                if self._docker_client:
                    self._container = self._docker_client.containers.run(
                        self._image_name,
                        command=enshell_cmd,
                        volumes=data_volume,
                        environment=container_env,
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                        user="ensight:ensight",
                    )
                logging.debug(f"_container = {str(self._container)}\n")
        else:
            if use_egl:  # pragma: no cover
                if self._docker_client:  # pragma: no cover
                    self._container = self._docker_client.containers.run(
                        self._image_name,
                        command=enshell_cmd,
                        environment=container_env,
                        device_requests=[
                            docker.types.DeviceRequest(count=-1, capabilities=[["gpu"]])
                        ],
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                        user="ensight:ensight",
                    )
            else:  # pragma: no cover
                logging.debug(
                    f"Running container {self._image_name} with cmd {enshell_cmd}\n"
                )  # pragma: no cover
                logging.debug(f"ports to map: {ports_to_map}\n")  # pragma: no cover
                if self._docker_client:  # pragma: no cover
                    self._container = self._docker_client.containers.run(  # pragma: no cover
                        self._image_name,
                        command=enshell_cmd,
                        environment=container_env,
                        ports=ports_to_map,
                        tty=True,
                        detach=True,
                        auto_remove=True,
                        remove=True,
                        user="ensight:ensight",
                    )
                # logging.debug(f"_container = {str(self._container)}\n")
        logging.debug("Container started.\n")
        return self.connect()

    def launch_webui(self, container_env_str):
        # Run websocketserver
        cmd = f"/ansys_inc/v{self._ansys_version}/FluidsOne/server/linx64/fluidsone"
        cmd += " --main-run-mode post"
        # websocket port - this needs to come first since we now have
        # --add_header as a optional arg that can take an arbitrary
        # number of optional headers.
        webui_port = self._service_host_port["webui"][1]
        grpc_port = self._service_host_port["grpc_private"][1]
        http_port = self._service_host_port["http"][1]
        ws_port = self._service_host_port["ws"][1]
        cmd += f" --server-listen-port {webui_port}"
        cmd += f" --server-web-roots /ansys_inc/v{self._ansys_version}/FluidsOne/web/ui"
        cmd += f" --ensight-grpc-port {grpc_port}"
        cmd += f" --ensight-html-port {http_port}"
        cmd += f" --ensight-ws-port {ws_port}"
        cmd += f" --ensight-session-directory {self._session_directory}"
        cmd += f" --ensight-secret-key {self._secret_key}"
        cmd += " --main-show-gui 'False'"
        logging.debug(f"Starting WebUI: {cmd}\n")
        ret = self._enshell.start_other(cmd, extra_env=container_env_str)
        if ret[0] != 0:  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError(f"Error starting WebUI: {cmd}\n")  # pragma: no cover

    def _get_build_info(self):
        # The unit test has no container
        if not self._container:
            return "mock"
        res = self._container.exec_run(["sh", "-lc", "ls -1 /ansys_inc 2>/dev/null"])
        entries = [
            e.strip() for e in res.output.decode("utf-8", "replace").splitlines() if e.strip()
        ]
        vdir = None
        for e in entries:
            if re.fullmatch(r"v\d{3}", e):
                vdir = e
                break
        path = f"/ansys_inc/{vdir}/CEI/BUILDINFO.txt"
        res2 = self._container.exec_run(["cat", path])
        return res2.output.decode("utf-8", errors="replace")

    def _grpc_version_check(self):
        text = self._get_build_info()
        if text == "mock":
            return True
        internal_version, ensight_full_version = self._get_versionfrom_buildinfo(text)
        return grpc_version_check(internal_version, ensight_full_version)

    def connect(self):
        """Create and bind a :class:`Session<ansys.pyensight.core.Session>` instance
        to the created EnSight gRPC connection started by EnShell.

        This method is an internal method.

        Returns
        -------
        obj
            :class:`Session<ansys.pyensight.core.Session>` instance

        Raises
        ------
        RuntimeError:
            Variety of error conditions.
        """
        self._has_grpc_changes = self._grpc_version_check()
        if not self._has_grpc_changes:
            warnings.warn(GRPC_WARNING_MESSAGE)
        #
        #
        # Start up the EnShell gRPC interface
        self._enshell = launch_enshell_interface(
            self._enshell_grpc_channel,
            self._service_host_port["grpc"][1],
            self._timeout,
            self._secret_key,
            grpc_allow_network_connections=self._grpc_allow_network_connections,
            grpc_disable_tls=self._grpc_disable_tls,
            grpc_uds_pathname=self._grpc_uds_pathname,
            grpc_use_tcp_sockets=self._grpc_use_tcp_sockets,
        )
        log_dir = "/data"
        if self._enshell_grpc_channel:  # pragma: no cover
            log_dir = "/work"

        if not self._enshell.is_connected():  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError("Can't connect to EnShell over gRPC.")  # pragma: no cover

        cmd = "set_no_reroute_log"
        ret = self._enshell.run_command(cmd)
        logging.debug(f"enshell cmd: {cmd} ret: {ret}\n")
        if ret[0] != 0:
            self.stop()
            raise RuntimeError(f"Error sending EnShell command: {cmd} ret: {ret}")

        files_to_try = [log_dir + "/enshell.log", "/home/ensight/enshell.log"]
        for f in files_to_try:  # pragma: no cover
            cmd = "set_debug_log " + f
            ret = self._enshell.run_command(cmd)
            if ret[0] == 0:
                self._enshell_log_file = f
                break
            else:
                logging.debug(f"enshell error; cmd: {cmd} ret: {ret}\n")

        if self._enshell_log_file is not None:  # pragma: no cover
            logging.debug(f"enshell log file {self._enshell_log_file}\n")

        cmd = "verbose 3"
        ret = self._enshell.run_command(cmd)
        logging.debug(f"enshell cmd: {cmd} ret: {ret}\n")
        if ret[0] != 0:  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError(
                f"Error sending EnShell command: {cmd} ret: {ret}"
            )  # pragma: no cover

        logging.debug("Connected to EnShell.  Getting CEI_HOME and Ansys version...\n")
        logging.debug(f"  _enshell: {self._enshell}\n\n")
        # Build up the command to run ensight via the EnShell gRPC interface

        self._cei_home = self._enshell.cei_home()
        self._ansys_version = self._enshell.ansys_version()
        print("CEI_HOME=", self._cei_home)
        print("Ansys Version=", self._ansys_version)

        logging.debug("Got them.  Starting EnSight...\n")

        use_egl = self._use_egl()

        # get the environment to pass to the container
        container_env_str = ""
        if self._pim_instance is not None:
            container_env = self._get_container_env()
            for i in container_env.items():
                container_env_str += f"{i[0]}={i[1]}\n"

        # Run EnSight
        ensight_env_vars = None
        if container_env_str != "":  # pragma: no cover
            ensight_env_vars = container_env_str  # pragma: no cover

        if use_egl:
            if ensight_env_vars is None:  # pragma: no cover
                ensight_env_vars = (
                    "LD_PRELOAD=/usr/local/lib64/libGL.so.1:/usr/local/lib64/libEGL.so.1"
                )
            else:
                ensight_env_vars += (  # pragma: no cover
                    "LD_PRELOAD=/usr/local/lib64/libGL.so.1:/usr/local/lib64/libEGL.so.1"
                )

        ensight_args = "-batch -v 3"

        if use_egl:
            ensight_args += " -egl"

        if self._use_sos:
            ensight_args += " -sos -nservers " + str(int(self._use_sos))

        ensight_args += " -grpc_server " + str(self._service_host_port["grpc_private"][1])

        if self._grpc_use_tcp_sockets:
            ensight_args += " -grpc_use_tcp_sockets"
        if self._grpc_allow_network_connections:
            ensight_args += " -grpc_allow_network_connections"
        if self._grpc_disable_tls:
            ensight_args += " -grpc_disable_tls"
        # Can't specify the same name; the default ought to be fine within the Docker Image
        # if self._grpc_uds_pathname:
        #    enshell_cmd += " -grpc_uds_pathname "+self._grpc_uds_pathname

        vnc_url = "vnc://%%3Frfb_port=1999%%26use_auth=0"
        ensight_args += " -vnc " + vnc_url
        if self._liben_rest:
            ensight_args += " -rest_server " + str(self._service_host_port["http"][1])
        if self._additional_command_line_options:
            ensight_args += " "
            ensight_args += " ".join(self._additional_command_line_options)

        logging.debug(f"Starting EnSight with args: {ensight_args}\n")
        if self._liben_rest:
            ensight_location = "/ansys_inc/v" + self._ansys_version + "/CEI/bin/ensight "
            ensight_args = ensight_location + ensight_args
            ret = self._enshell.start_other(ensight_args, extra_env=ensight_env_vars)
        else:
            ret = self._enshell.start_ensight(ensight_args, ensight_env_vars)
        if ret[0] != 0:  # pragma: no cover
            self.stop()  # pragma: no cover
            raise RuntimeError(
                f"Error starting EnSight with args: {ensight_args}"
            )  # pragma: no cover

        logging.debug("EnSight started.  Starting wss...\n")

        # Run websocketserver
        if not self._liben_rest:
            logging.debug(
                "WebSocketserver script not being launched. WS server must be launched manually.\n\n"
            )
            wss_cmd = "cpython /ansys_inc/v" + self._ansys_version + "/CEI/nexus"
            wss_cmd += self._ansys_version + "/nexus_launcher/websocketserver.py"
            # websocket port - this needs to come first since we now have
            # --add_header as a optional arg that can take an arbitrary
            # number of optional headers.
            if int(self._ansys_version) > 252 and self._do_not_start_ws:
                wss_cmd += " -1"
            else:
                wss_cmd += " " + str(self._service_host_port["ws"][1])
            #
            wss_cmd += " --http_directory " + self._session_directory
            # http port
            wss_cmd += " --http_port " + str(self._service_host_port["http"][1])
            # vnc port
            if int(self._ansys_version) > 252 and self._rest_ws_separate_loops:
                wss_cmd += " --separate_loops"
            wss_cmd += f" --security_token {self._secret_key}"
            wss_cmd += " --client_port 1999"
            # optional PIM instance header
            if self._pim_instance is not None:
                # Add the PIM instance header. wss needs to return this optional
                # header in each http response.  It's how the Ansys Lab proxy
                # knows how to map back to this particular container's IP and port.
                wss_cmd += " --add_header instance_name=" + self._pim_instance.name
            # EnSight REST API
            if self._enable_rest_api:
                # grpc port
                wss_cmd += " --grpc_port " + str(self._service_host_port["grpc_private"][1])
                if self._grpc_use_tcp_sockets:
                    wss_cmd += " --grpc_use_tcp_sockets"
                if self._grpc_allow_network_connections:
                    wss_cmd += " --grpc_allow_network_connections"
                if self._grpc_disable_tls:
                    wss_cmd += " --grpc_disable_tls"
                if self._grpc_uds_pathname:
                    wss_cmd += " --grpc_uds_pathname"
                    wss_cmd += " " + self._grpc_uds_pathname
            # EnVision sessions
            wss_cmd += " --local_session envision 5"

            wss_env_vars = None
            if container_env_str != "":  # pragma: no cover
                wss_env_vars = container_env_str  # pragma: no cover

            logging.debug(f"Starting WSS: {wss_cmd}\n")
            ret = self._enshell.start_other(wss_cmd, extra_env=wss_env_vars)
            if ret[0] != 0:  # pragma: no cover
                self.stop()  # pragma: no cover
                raise RuntimeError(f"Error starting WSS: {wss_cmd}\n")  # pragma: no cover

            logging.debug("wss started.  Making session...\n")

        # build the session instance
        # WARNING: assuming the host is the same for grpc_private, http, and ws
        # This may not be true in the future if using PIM.
        # revise Session to handle three different hosts if necessary.
        use_sos = False
        if self._use_sos:
            use_sos = True
        if self._pim_instance is None:
            ws_port = self._service_host_port["ws"][1]
        else:
            ws_port = self._service_host_port["http"][1]  # pragma: no cover
        session = ansys.pyensight.core.session.Session(
            host=self._service_host_port["grpc_private"][0],
            grpc_port=self._service_host_port["grpc_private"][1],
            grpc_use_tcp_sockets=self._grpc_use_tcp_sockets,
            grpc_allow_network_connections=self._grpc_allow_network_connections,
            grpc_disable_tls=self._grpc_disable_tls,
            grpc_uds_pathname=self._grpc_uds_pathname,
            html_hostname=self._service_host_port["http"][0],
            html_port=self._service_host_port["http"][1],
            ws_port=ws_port,
            install_path=None,
            secret_key=self._secret_key,
            timeout=self._timeout,
            sos=use_sos,
            rest_api=self._enable_rest_api,
            webui_port=self._service_host_port["webui"][1] if self._launch_webui else None,
        )
        session.launcher = self
        self._sessions.append(session)

        if self._launch_webui:
            self.launch_webui(container_env_str)
        logging.debug("Return session.\n")

        return session

    def close(self, session):
        """Shut down the launched EnSight session.

        This method closes all associated sessions and then stops the
        launched EnSight instance.

        Parameters
        ----------
        session : ``pyensight.Session``
            Session to close.

        Raises
        ------
        RuntimeError
            If the session was not launched by this launcher.

        """
        if self._enshell:
            if self._enshell.is_connected():  # pragma: no cover
                logging.debug("Killing WSS\n")
                command = 'pkill -f "websocketserver.py"'
                kill_env_vars = None
                container_env_str = ""
                if self._pim_instance is not None:
                    container_env = self._get_container_env()
                    for i in container_env.items():
                        container_env_str += f"{i[0]}={i[1]}\n"
                if container_env_str:  # pragma: no cover
                    kill_env_vars = container_env_str  # pragma: no cover
                ret = self._enshell.start_other(command, extra_env=kill_env_vars)
                if ret[0] != 0:  # pragma: no cover
                    pass
        return super().close(session)

    def stop(self) -> None:
        """Release any additional resources allocated during launching."""
        if self._enshell:
            if self._enshell.is_connected():  # pragma: no cover
                try:
                    logging.debug("Killing WSS\n")
                    command = 'pkill -f "websocketserver.py"'
                    kill_env_vars = None
                    container_env_str = ""
                    if self._pim_instance is not None:
                        container_env = self._get_container_env()
                        for i in container_env.items():
                            container_env_str += f"{i[0]}={i[1]}\n"
                    if container_env_str:  # pragma: no cover
                        kill_env_vars = container_env_str  # pragma: no cover
                    ret = self._enshell.start_other(command, extra_env=kill_env_vars)
                    if ret[0] != 0:  # pragma: no cover
                        pass
                    logging.debug("Stopping EnShell.\n")
                    self._enshell.stop_server()
                except Exception:  # pragma: no cover
                    pass  # pragma: no cover
                self._enshell = None
        #
        if self._container:
            try:
                logging.debug("Stopping the Docker Container.\n")
                self._container.stop()
            except Exception:
                pass
            try:
                logging.debug("Removing the Docker Container.\n")
                self._container.remove(force=True)
            except Exception:
                pass
            self._container = None

        if self._pim_instance is not None:
            logging.debug("Deleting the PIM instance.\n")
            self._pim_instance.delete()
            self._pim_instance = None
        super().stop()

    def file_service(self) -> Optional[Any]:
        """Get the PIM file service object if available."""
        return self._pim_file_service

    def _is_system_egl_capable(self) -> bool:
        """Check if the system is EGL capable.

        Returns
        -------
        bool
            ``True`` if the system is EGL capable, ``False`` otherwise.
        """
        if self._is_windows():
            return False  # pragma: no cover

        # FIXME: MFK, need to figure out how we'd do this
        # with a PIM managed system such as Ansys Lab
        if self._pim_instance is not None:
            return False

        try:  # pragma: no cover
            subprocess.check_output("nvidia-smi")
            return True
        except (subprocess.CalledProcessError, FileNotFoundError):
            return False

    def enshell_log_contents(self) -> Optional[str]:
        """Get the contents of the EnShell log if possible.

        Returns
        -------
        str
           Contents of the log or ``None``.
        """
        if self._enshell_log_file is None:  # pragma: no cover
            return None  # pragma: no cover

        if self._container is not None:
            try:
                # docker containers allow copying from a container to the host
                # a file, files, or a directory.  it arrives as a tar file.
                # we're grabbing just the log file, but we still need to
                # extract it from the tar file and then put the contents
                # in the string we're returning
                from io import BytesIO
                import tarfile

                bits, stat = self._container.get_archive(self._enshell_log_file)
                file_obj = BytesIO()
                for chunk in bits:
                    file_obj.write(chunk)
                file_obj.seek(0)
                tar = tarfile.open(mode="r", fileobj=file_obj)
                member = tar.getmembers()
                text = tar.extractfile(member[0])
                s = text.read().decode("utf-8")
                text.close()
                return s
            except Exception as e:  # pragma: no cover
                logging.debug(f"Error getting EnShell log: {e}\n")  # pragma: no cover
                return None  # pragma: no cover

        fs = self.file_service()  # pragma: no cover
        if fs is not None:  # pragma: no cover
            try:
                fs.download_file("enshell.log", ".")
                f = open("enshell.log")
                s = f.read()
                f.close()
                return s
            except Exception:
                return None
